"""
:mod:`orion.serving.parameters` -- Common code for verifying query parameters
=============================================================================

.. module:: parameters
   :platform: Unix
   :synopsis: Common code related to query parameters verification
"""
from typing import Optional

import falcon

from orion.core.io import experiment_builder
from orion.core.utils.exceptions import NoConfigurationError
from orion.core.worker.experiment import Experiment
from orion.core.worker.trial import Trial
from orion.serving.responses import ERROR_EXPERIMENT_NOT_FOUND, ERROR_INVALID_PARAMETER, \
    ERROR_TRIAL_NOT_FOUND


def verify_query_parameters(parameters: dict, supported_parameters: list):
    """
    Verifies that the parameters given in the input dictionary are all supported.

    Parameter
    ---------
    parameters
        The dictionary of parameters to verify in the format ``parameter_name:value``.
    supported_parameters
        The list of parameters that are supported.

    Raises
    ------
    falcon.HTTPBadRequest
        When a parameter is not listed in the supported parameters.
    """
    for parameter in parameters:
        if parameter not in supported_parameters:
            description = _compose_error_message(parameter, supported_parameters)
            raise falcon.HTTPBadRequest(ERROR_INVALID_PARAMETER, description)


def verify_status(status):
    """Verifies that the given trial status is supported. Raises falcon.HTTPBadRequest otherwise"""
    if status and status not in Trial.allowed_stati:
        description = 'The "status" parameter is invalid. '
        description += 'The value of the parameter must be one of {}'.format(
            list(Trial.allowed_stati))

        raise falcon.HTTPBadRequest(ERROR_INVALID_PARAMETER, description)


def _compose_error_message(key: str, supported_parameters: list):
    """Creates the error message depending on the number of supported parameters available."""
    error_message = f'Parameter "{key}" is not supported. Expected '

    if len(supported_parameters) > 1:
        supported_parameters.sort()
        error_message += f'one of {supported_parameters}.'
    else:
        error_message += f'parameter "{supported_parameters[0]}".'

    return error_message


def retrieve_experiment(experiment_name: str, version: int = None) -> Optional[Experiment]:
    """
    Retrieve an experiment from the database with the given name and version.

    Raises
    ------
    falcon.HTTPNotFound
        When the experiment doesn't exist
    """
    try:
        experiment = experiment_builder.build_view(experiment_name, version)
        if version and experiment.version != version:
            raise falcon.HTTPNotFound(
                title=ERROR_EXPERIMENT_NOT_FOUND,
                description=f'Experiment "{experiment_name}" has no version "{version}"')
        return experiment
    except NoConfigurationError:
        raise falcon.HTTPNotFound(title=ERROR_EXPERIMENT_NOT_FOUND,
                                  description=f'Experiment "{experiment_name}" does not exist')


def retrieve_trial(experiment: Experiment, trial_id: str):
    """
    Retrieves the trial for the given id in the experiment

    Parameters
    ----------
    experiment: Experiment
        The experiment containing the trial.

    trial_id: str
        The id of the trial

    Raises
    ------
    falcon.HTTPNotFound
        When the trial doesn't exist in the experiment.

    """
    trial = experiment.get_trial(uid=trial_id)
    if not trial:
        raise falcon.HTTPNotFound(title=ERROR_TRIAL_NOT_FOUND,
                                  description=f'Trial "{trial_id}" does not exist')
    return trial



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Installation script for `orion.algo.gradient_descent`."""
from setuptools import setup

setup_args = dict(
    name='orion.algo.gradient_descent',
    version=0.1,
    description="Implement a gradient descent algorithm, for demo and testing.",
    license='BSD-3-Clause',
    author='Christos Tsirigotis',
    author_email='tsirif@gmail.com',
    url='https://github.com/epistimio/orion',
    packages=['orion.algo'],
    package_dir={'': 'src'},
    include_package_data=True,
    entry_points={
        'OptimizationAlgorithm': [
            'gradient_descent = orion.algo.gradient_descent:Gradient_Descent'
            ],
        },
    install_requires=['orion'],
    setup_requires=['setuptools'],
    # "Zipped eggs don't play nicely with namespace packaging"
    # from https://github.com/pypa/sample-namespace-packages
    zip_safe=False
    )

setup_args['keywords'] = [
    'Machine Learning',
    'Deep Learning',
    'Distributed',
    'Optimization',
    ]

setup_args['platforms'] = ['Linux']

setup_args['classifiers'] = [
    'Development Status :: 1 - Planning',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
] + [('Programming Language :: Python :: %s' % x)
     for x in '3 3.6 3.7 3.8'.split()]

if __name__ == '__main__':
    setup(**setup_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.terminal` -- Helper functions for terminal i/o
=====================================================================

.. module:: setup
   :platform: Unix
   :synopsis: Helper functions for interaction with user on terminal
"""


def ask_question(question, default=None):
    """Ask a question to the user and receive an answer.

    Parameters
    ----------
    question: str
        The question to be asked.
    default: str
        The default value to use if the user enters nothing.

    Returns
    -------
    str
        The answer provided by the user.

    """
    if default is not None:
        question = question + " (default: {}) ".format(default)

    answer = input(question)

    if answer.strip() == "":
        return default

    return answer


def confirm_name(message, name, force=False):
    """Ask the user to confirm the name.

    Parameters
    ----------
    message: str
        The message to be printed.
    name: str
        The string that the user must enter.
    force: bool
        Override confirmation and return True. Default: False.

    Returns
    -------
    bool
        True if confirmed, False otherwise.

    """
    if force:
        print(message)
        print('FORCED')
        return True

    answer = input(message)

    return answer.strip() == name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for algos included with orion."""
import copy
import random

import pytest

from orion.client import workon


storage = {
    'type': 'legacy',
    'database': {'type': 'ephemeraldb'}}


space = {
    'x': 'uniform(-50, 50)'
    }


space_with_fidelity = {
    'x': space['x'],
    'noise': 'fidelity(1,10,4)'}


algorithm_configs = {
    'random': {
        'random': {
            'seed': 1}},
    'tpe': {
        'tpe': {
            'seed': 1,
            'n_initial_points': 20,
            'n_ei_candidates': 24,
            'gamma': 0.25,
            'equal_weight': False,
            'prior_weight': 1.0,
            'full_weight_num': 25}},
    'asha': {
        'asha': {
            'seed': 1,
            'num_rungs': 4,
            'num_brackets': 1,
            'grace_period': None,
            'max_resources': None,
            'reduction_factor': None}},
    'hyperband': {
        'hyperband': {
            'repetitions': 5,
            'seed': 1}}
    }

no_fidelity_algorithms = ['random', 'tpe']
no_fidelity_algorithm_configs = {key: algorithm_configs[key] for key in no_fidelity_algorithms}

fidelity_only_algorithms = ['asha', 'hyperband']
fidelity_only_algorithm_configs = {key: algorithm_configs[key] for key in fidelity_only_algorithms}


def rosenbrock(x, noise=None):
    """Evaluate partial information of a quadratic."""
    z = (x - 34.56789)
    if noise:
        noise = (1 - noise / 10) + 0.0001
        z *= random.gauss(0, noise)

    return [{'name': 'objective', 'type': 'objective', 'value': 4 * z**2 + 23.4},
            {'name': 'gradient', 'type': 'gradient', 'value': [8 * z]}]


@pytest.mark.parametrize(
    'algorithm',
    fidelity_only_algorithm_configs.values(),
    ids=list(fidelity_only_algorithm_configs.keys()))
def test_missing_fidelity(algorithm):
    """Test a simple usage scenario."""
    with pytest.raises(RuntimeError) as exc:
        workon(rosenbrock, space, algorithms=algorithm, max_trials=100)

    assert "https://orion.readthedocs.io/en/develop/user/algorithms.html" in str(exc.value)


@pytest.mark.parametrize(
    'algorithm',
    no_fidelity_algorithm_configs.values(),
    ids=list(no_fidelity_algorithm_configs.keys()))
def test_simple(algorithm):
    """Test a simple usage scenario."""
    max_trials = 100
    exp = workon(rosenbrock, space, algorithms=algorithm, max_trials=max_trials)

    assert exp.max_trials == max_trials
    assert exp.configuration['algorithms'] == algorithm

    trials = exp.fetch_trials()
    assert len(trials) == max_trials
    assert trials[-1].status == 'completed'

    best_trial = sorted(trials, key=lambda trial: trial.objective.value)[0]
    assert best_trial.objective.name == 'objective'
    assert abs(best_trial.objective.value - 23.4) < 0.1
    assert len(best_trial.params) == 1
    param = best_trial._params[0]
    assert param.name == 'x'
    assert param.type == 'real'


@pytest.mark.parametrize(
    'algorithm',
    no_fidelity_algorithm_configs.values(),
    ids=list(no_fidelity_algorithm_configs.keys()))
def test_cardinality_stop(algorithm):
    """Test when algo needs to stop because all space is explored (dicrete space)."""
    discrete_space = copy.deepcopy(space)
    discrete_space['x'] = 'uniform(-10, 5, discrete=True)'
    exp = workon(rosenbrock, discrete_space, algorithms=algorithm, max_trials=100)

    trials = exp.fetch_trials()
    assert len(trials) == 15
    assert trials[-1].status == 'completed'


@pytest.mark.parametrize(
    'algorithm',
    algorithm_configs.values(),
    ids=list(algorithm_configs.keys()))
def test_with_fidelity(algorithm):
    """Test a scenario with fidelity."""
    exp = workon(rosenbrock, space_with_fidelity, algorithms=algorithm, max_trials=100)

    assert exp.configuration['algorithms'] == algorithm

    trials = exp.fetch_trials()
    assert len(trials) <= 100
    assert trials[-1].status == 'completed'

    results = [trial.objective.value for trial in trials]
    print(min(results))
    print(max(results))
    best_trial = next(iter(sorted(trials, key=lambda trial: trial.objective.value)))

    assert best_trial.objective.name == 'objective'
    assert abs(best_trial.objective.value - 23.4) < 1e-5
    assert len(best_trial.params) == 2
    fidelity = best_trial._params[0]
    assert fidelity.name == 'noise'
    assert fidelity.type == 'fidelity'
    assert fidelity.value == 10
    param = best_trial._params[1]
    assert param.name == 'x'
    assert param.type == 'real'



# -*- coding: utf-8 -*-
"""
:mod:`orion.serving.responses` -- Helpers for building responses according to the specification
===============================================================================================

.. module:: responses
   :platform: Unix
   :synopsis: Offers functions and attributes to generate response objects according
      to the API specification
"""
from orion.core.worker.experiment import Experiment
from orion.core.worker.trial import Trial

ERROR_EXPERIMENT_NOT_FOUND = 'Experiment not found'
ERROR_INVALID_PARAMETER = 'Invalid parameter'
ERROR_TRIAL_NOT_FOUND = 'Trial not found'


def build_trial_response(trial: Trial) -> dict:
    """
    Build the response representing a trial according to the API specification.

    Parameters
    ----------
    trial: Trial
        The trial to return for the API.

    Returns
    -------
    A JSON-serializable dict representing the given trial.

    """
    return {'id': trial.id,
            'submitTime': str(trial.submit_time),
            'startTime': str(trial.start_time),
            'endTime': str(trial.end_time),
            'parameters': trial.params,
            'objective': trial.objective.value,
            'statistics': {statistic.name: statistic.value for statistic in trial.statistics}}


def build_experiment_response(experiment: Experiment,
                              status: str,
                              algorithm: dict,
                              best_trial: Trial = None):
    """
    Build the response representing an experiment response object according to the API
    specification.

    Parameters
    ----------
    experiment: Experiment
        The experiment to return to the API
    status: str
        The status of the experiment
    algorithm: dict
        The dictionary containing the algorithm's configuration
    best_trial: Trial (Optional)
        The best trial to date of the experiment

    Returns
    -------
    A JSON-serializable experiment response object representing the given experiment.
    """
    return {
        "name": experiment.name,
        "version": experiment.version,
        "status": status,
        "trialsCompleted": experiment.stats['trials_completed'] if experiment.stats else 0,
        "startTime": str(experiment.stats['start_time']) if experiment.stats else None,
        "endTime": str(experiment.stats['finish_time']) if experiment.stats else None,
        "user": experiment.metadata['user'],
        "orionVersion": experiment.metadata['orion_version'],
        "config": {
            "maxTrials": experiment.max_trials,
            "poolSize": experiment.pool_size,
            "algorithm": algorithm,
            "space": experiment.configuration['space']
        },
        "bestTrial": build_trial_response(best_trial) if best_trial else {}
    }


def build_experiments_response(experiments: dict):
    """
    Build the response representing a list of experiments according to the API specification.

    Parameters
    ----------
    experiments: dict
        A dict containing pairs of ``experiment-name:experiment-version``.

    Returns
    -------
    A JSON-serializable list of experiments as defined in the API specification.
    """
    result = []
    for name, version in experiments.items():
        result.append({
            'name': name,
            'version': version
        })
    return result


def build_trials_response(trials: list):
    """
    Build the response representing a list of trials according to the API specification.

    Parameters
    ----------
    trials: list
        A list of :class:`orion.core.worker.trial.Trial`.

    Returns
    -------
    A JSON-serializable list of trials as defined in the API specification.
    """
    response = []
    for trial in trials:
        response.append({
            'id': trial.id
        })
    return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.algo.space`."""

from collections import (defaultdict, OrderedDict)
import sys

import numpy as np
from numpy.testing import assert_array_equal as assert_eq
import pytest
from scipy.stats import distributions as dists

from orion.algo.space import (
    Categorical, check_random_state, Dimension, Fidelity, Integer, Real, Space)


class TestCheckRandomState:
    """Test `orion.algo.space.check_random_state`"""

    def test_rng_null(self):
        """Test that passing None returns numpy._rand"""
        assert check_random_state(None) is np.random.mtrand._rand

    def test_rng_random_state(self):
        """Test that passing RandomState returns itself"""
        rng = np.random.RandomState(1)
        assert check_random_state(rng) is rng

    def test_rng_int(self):
        """Test that passing int returns RandomState"""
        rng = check_random_state(1)
        assert isinstance(rng, np.random.RandomState)
        assert rng is not np.random.mtrand._rand

    def test_rng_tuple(self):
        """Test that passing int returns RandomState"""
        rng = check_random_state((1, 12, 123))
        assert isinstance(rng, np.random.RandomState)
        assert rng is not np.random.mtrand._rand

    def test_rng_invalid_value(self):
        """Test that passing int returns RandomState"""
        with pytest.raises(ValueError) as exc:
            check_random_state('oh_no_oh_no')

        assert '\'oh_no_oh_no\' cannot be used to seed' in str(exc.value)


class TestDimension(object):
    """Test methods of a Dimension object."""

    def test_simple_instance(self, seed):
        """Test Dimension.__init__."""
        dim = Dimension('yolo', 'norm', 0.9, 0.1)
        samples = dim.sample(seed=seed)
        assert len(samples) == 1
        assert dists.norm.rvs(0.9, 0.1) == samples[0]

        assert dists.norm.interval(1.0, 0.9, 0.1) == dim.interval()
        assert dists.norm.interval(0.5, 0.9, 0.1) == dim.interval(0.5)

        assert str(dim) == "Dimension(name=yolo, prior={norm: (0.9, 0.1), {}}, " \
                           "shape=(), default value=None)"

        assert dim.name == 'yolo'
        assert dim.type == 'dimension'
        assert dim.shape == ()

    def test_shaped_instance(self, seed):
        """Use shape keyword argument."""
        dim = Dimension('yolo', 'norm', 0.9, shape=(3, 2))
        samples = dim.sample(seed=seed)
        assert len(samples) == 1
        assert_eq(dists.norm.rvs(0.9, size=(3, 2)), samples[0])

        assert dim.shape == (3, 2)

        dim = Dimension('yolo', 'norm', 0.9, shape=4)
        samples = dim.sample(seed=seed)
        assert len(samples) == 1
        assert_eq(dists.norm.rvs(0.9, size=4), samples[0])

        assert dim.shape == (4,)

    def test_ban_size_kwarg(self):
        """Should not be able to use 'size' kwarg."""
        with pytest.raises(ValueError):
            Dimension('yolo', 'norm', 0.9, size=(3, 2))

    def test_ban_seed_kwarg(self):
        """Should not be able to use 'seed' kwarg."""
        with pytest.raises(ValueError):
            Dimension('yolo', 'norm', 0.9, seed=8)

    def test_ban_rng_kwarg(self):
        """Should not be able to use 'random_state' kwarg."""
        with pytest.raises(ValueError):
            Dimension('yolo', 'norm', 0.9, random_state=8)

    def test_with_predefined_dist(self, seed):
        """Use an already defined distribution object as prior arg."""
        dim = Dimension('yolo', dists.norm, 0.9)
        samples = dim.sample(seed=seed)
        assert len(samples) == 1
        assert dists.norm.rvs(0.9) == samples[0]

    def test_ban_discrete_kwarg(self):
        """Do not allow use for 'discrete' kwarg, because now there's `_Discrete`."""
        with pytest.raises(ValueError) as exc:
            Dimension('yolo', 'uniform', -3, 4, shape=(4, 4), discrete=True)
        assert "pure `_Discrete`" in str(exc.value)

    def test_many_samples(self, seed):
        """More than 1."""
        dim = Dimension('yolo', 'uniform', -3, 4, shape=(4, 4))
        samples = dim.sample(n_samples=4, seed=seed)
        assert len(samples) == 4
        assert_eq(dists.uniform.rvs(-3, 4, size=(4, 4)), samples[0])

    def test_interval(self):
        """Test that bounds on variable."""
        dim = Dimension('yolo', 'uniform', -3, 4)
        assert dim.interval(1.0) == (-3.0, 1.0)  # reminder that `scale` is not upper bound

    def test_contains_bounds(self):
        """Test __contains__ for bounds."""
        dim = Dimension('yolo', 'uniform', -3, 4)
        with pytest.raises(NotImplementedError):
            assert -3 in dim

    def test_contains_shape(self):
        """Test __contains__ for shape check."""
        dim = Dimension(None, 'uniform', -3, 4, shape=(4, 4))

        with pytest.raises(NotImplementedError):
            assert dists.uniform.rvs(-3, 4, size=(4, 4)) in dim

    def test_set_bad_name(self):
        """Try setting a name other than str or None."""
        dim = Dimension('yolo', 'uniform', -3, 4, shape=(4, 4))
        with pytest.raises(TypeError):
            dim.name = 4

    def test_init_with_default_value(self):
        """Make sure the __contains__ method does not work"""
        with pytest.raises(NotImplementedError):
            Dimension('yolo', 'uniform', -3, 4, default_value=4)

    def test_no_default_value(self):
        """Test that no giving a default value assigns None"""
        dim = Dimension('yolo', 'uniform', -3, 4)
        assert dim.default_value is None

    def test_no_prior(self):
        """Test that giving a null prior defaults prior_name to `None`."""
        dim = Dimension('yolo', None)
        print(dim._prior_name)
        assert dim.prior is None
        assert dim._prior_name == 'None'

    @pytest.mark.skipif(sys.version_info < (3, 6), reason="requires python3.6 or higher")
    def test_get_prior_string(self):
        """Test that prior string can be rebuilt."""
        dim = Dimension('yolo', 'alpha', 1, 2, 3, some='args', plus='fluff', n=4)
        assert dim.get_prior_string() == 'alpha(1, 2, 3, some=\'args\', plus=\'fluff\', n=4)'

    def test_get_prior_string_uniform(self):
        """Test special uniform args are handled properly."""
        dim = Dimension('yolo', 'uniform', 1, 2)
        assert dim.get_prior_string() == 'uniform(1, 3)'

    def test_get_prior_string_default_values(self, monkeypatch):
        """Test that default_value are included."""
        def contains(self, value):
            return True
        monkeypatch.setattr(Dimension, '__contains__', contains)
        dim = Dimension('yolo', 'alpha', 1, 2, default_value=1)
        assert dim.get_prior_string() == 'alpha(1, 2, default_value=1)'

    def test_get_prior_string_shape(self):
        """Test that shape is included."""
        dim = Dimension('yolo', 'alpha', 1, 2, shape=(2, 3))
        assert dim.get_prior_string() == 'alpha(1, 2, shape=(2, 3))'

    def test_get_prior_string_loguniform(self):
        """Test that special loguniform prior name is replaced properly."""
        dim = Dimension('yolo', 'reciprocal', 1e-10, 1)
        assert dim.get_prior_string() == 'loguniform(1e-10, 1)'

    def test_prior_name(self):
        """Test prior name is correct in dimension"""
        dim = Dimension('yolo', 'reciprocal', 1e-10, 1)
        assert dim.prior_name == 'reciprocal'

        dim = Dimension('yolo', 'norm', 0.9)
        assert dim.prior_name == 'norm'

        dim = Real('yolo', 'uniform', 1, 2)
        assert dim.prior_name == 'uniform'

        dim = Integer('yolo1', 'uniform', -3, 6)
        assert dim.prior_name == 'int_uniform'

        dim = Integer('yolo1', 'norm', -3, 6)
        assert dim.prior_name == 'int_norm'

        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 'lalala': 0.4}
        dim = Categorical('yolo', categories)
        assert dim.prior_name == 'choices'


class TestReal(object):
    """Test methods of a `Real` object."""

    def test_simple_instance(self, seed):
        """Test Real.__init__."""
        dim = Real('yolo', 'norm', 0.9)
        samples = dim.sample(seed=seed)
        assert len(samples) == 1
        assert dists.norm.rvs(0.9) == samples[0]

        assert dists.norm.interval(1.0, 0.9) == dim.interval()
        assert dists.norm.interval(0.5, 0.9) == dim.interval(0.5)

        assert 1.0 in dim

        assert str(dim) == "Real(name=yolo, prior={norm: (0.9,), {}}, shape=(), default value=None)"
        assert dim.name == 'yolo'
        assert dim.type == 'real'
        assert dim.shape == ()

    def test_contains_extra_bounds(self):
        """Test __contains__ for the extra bounds."""
        dim = Real('yolo', 'norm', 0, 3, low=-3, high=+3)
        assert dists.uniform.rvs(-3, 3) in dim
        assert -4 not in dim
        assert +4 not in dim
        assert (1, 2) not in dim

    def test_sample_from_extra_bounds_good(self):
        """Randomized test **successful** sampling with the extra bounds."""
        dim = Real('yolo', 'norm', 0, 2, low=-5, high=+5, shape=(4, 4))
        for _ in range(8):
            samples = dim.sample(8)
            for sample in samples:
                assert sample in dim

    def test_sample_from_extra_bounds_bad(self):
        """Randomized test **unsuccessfully** sampling with the extra bounds."""
        dim = Real('yolo', 'norm', 0, 2, low=-2, high=+2, shape=(4, 4))
        with pytest.raises(ValueError) as exc:
            dim.sample(8)
        assert "Improbable bounds" in str(exc.value)

    def test_bad_bounds(self):
        """Try setting bound with high <= low."""
        with pytest.raises(ValueError):
            Real('yolo', 'norm', 0, 2, low=+2, high=-2, shape=(4, 4))
        with pytest.raises(ValueError):
            Real('yolo', 'norm', 0, 2, low=+2, high=+2, shape=(4, 4))

    def test_interval(self):
        """Interval takes into account explicitly bounds."""
        dim = Real('yolo', 'norm', 0, 3, low=-3, high=+3)
        assert dim.interval() == (-3, 3)

        dim = Real('yolo', 'alpha', 0.9, low=-3, high=+3)
        assert dim.interval() == (0, 3)

        dim = Real('yolo', 'uniform', -2, 4, low=-3, high=+3)
        assert dim.interval() == (-2.0, 2.0)

    def test_init_with_default_value(self):
        """Make sure the default value is set"""
        dim = Real('yolo', 'uniform', -3, 10, default_value=2.0)

        assert type(dim.default_value) is float

    def test_set_outside_bounds_default_value(self):
        """Make sure default value is inside the bounds"""
        with pytest.raises(ValueError):
            Real('yolo', 'uniform', -3, 2, default_value=5)

    def test_no_default_value(self):
        """Make sure the default value is None"""
        dim = Real('yolo', 'uniform', -3, 4)
        assert dim.default_value is None

    def test_cast_list(self):
        """Make sure list are cast to float and returned as list of values"""
        dim = Real('yolo', 'uniform', -3, 4)
        assert dim.cast(['1', '2']) == [1.0, 2.0]

    def test_cast_array(self):
        """Make sure array are cast to float and returned as array of values"""
        dim = Real('yolo', 'uniform', -3, 4)
        assert np.all(dim.cast(np.array(['1', '2'])) == np.array([1.0, 2.0]))


class TestInteger(object):
    """Test methods of a `Integer` object."""

    def test_simple_instance(self, seed):
        """Test Integer.__init__."""
        dim = Integer('yolo', 'uniform', -3, 6)
        samples = dim.sample(seed=seed)
        assert len(samples) == 1
        assert samples[0] == -2

        assert dim.interval() == (-3, 3)
        assert dim.interval(0.5) == (-2, 2)

        assert 1.0 in dim

        assert str(dim) == "Integer(name=yolo, prior={uniform: (-3, 6), {}}, "\
                           "shape=(), default value=None)"

        assert dim.name == 'yolo'
        assert dim.type == 'integer'
        assert dim.shape == ()

    def test_inclusive_intervals(self):
        """Test that discretized bounds are valid"""
        dim = Integer('yolo', 'uniform', -3, 5.5)
        assert dim.interval() == (-3, 3)

    def test_contains(self):
        """Check for integer test."""
        dim = Integer('yolo', 'uniform', -3, 6)

        assert 0.1 not in dim
        assert (0.1, -0.2) not in dim
        assert 0 in dim
        assert (1, 2) not in dim
        assert 6 not in dim
        assert -3 in dim
        assert -4 not in dim

    def test_interval_with_infs(self):
        """Regression test: Interval handles correctly extreme bounds."""
        dim = Integer('yolo', 'poisson', 5)
        # XXX: Complete this on both end of interval when scipy bug is fixed
        assert dim.interval()[1] == np.inf

    @pytest.mark.xfail(reason="scipy bug")
    def test_scipy_integer_dist_interval_bug(self):
        """Scipy does not return the correct answer for integer distributions."""
        dim = Integer('yolo', 'randint', -3, 6)
        assert dim.interval() == (-3, 6)
        assert dim.interval(1.0) == (-3, 6)
        assert dim.interval(0.9999) == (-3, 6)

        dim = Integer('yolo2', 'randint', -2, 4, loc=8)
        assert dim.interval() == (6, 12)

    def test_init_with_default_value(self):
        """Make sure the type of the default value is int"""
        dim = Integer('yolo', 'uniform', -3, 10, default_value=2)

        assert type(dim.default_value) is int

    def test_set_outside_bounds_default_value(self):
        """Make sure the default value is inside the bounds of the dimensions"""
        with pytest.raises(ValueError):
            Integer('yolo', 'uniform', -3, 2, default_value=4)

    def test_no_default_value(self):
        """Make sure the default value is None"""
        dim = Integer('yolo', 'uniform', -3, 4)
        assert dim.default_value is None

    def test_cast_list(self):
        """Make sure list are cast to int and returned as list of values"""
        dim = Integer('yolo', 'uniform', -3, 4)
        assert dim.cast(['1', '2']) == [1, 2]

    def test_cast_array(self):
        """Make sure array are cast to int and returned as array of values"""
        dim = Integer('yolo', 'uniform', -3, 4)
        assert np.all(dim.cast(np.array(['1', '2'])) == np.array([1, 2]))

    def test_get_prior_string_discrete(self):
        """Test that discrete is included."""
        dim = Integer('yolo', 'uniform', 1, 2)
        assert dim.get_prior_string() == 'uniform(1, 3, discrete=True)'


class TestCategorical(object):
    """Test methods of a `Categorical` object."""

    def test_with_tuple(self, seed):
        """Test Categorical.__init__ with a tuple."""
        categories = ('asdfa', 2)
        dim = Categorical('yolo', categories)
        samples = dim.sample(seed=seed)
        assert len(samples) == 1
        assert samples[0] == 'asdfa'
        assert dim._probs == (0.5, 0.5)

        assert categories == dim.categories

        assert 2 in dim
        assert 3 not in dim

        assert str(dim) == "Categorical(name=yolo, prior={asdfa: 0.50, 2: 0.50}, "\
                           "shape=(), default value=None)"

        assert dim.name == 'yolo'
        assert dim.type == 'categorical'
        assert dim.shape == ()

    def test_with_dict(self, seed):
        """Test Categorical.__init__ with a dictionary."""
        probs = (0.1, 0.2, 0.3, 0.4)
        categories = ('asdfa', 2, 3, 4)
        dim = Categorical('yolo', OrderedDict(zip(categories, probs)))
        samples = dim.sample(seed=seed)
        assert len(samples) == 1
        assert samples[0] == 2
        assert dim._probs == probs

        assert categories == dim.categories

        assert 2 in dim
        assert 0 not in dim

        assert dim.name == 'yolo'
        assert dim.type == 'categorical'
        assert dim.shape == ()

    def test_probabilities_are_ok(self, seed):
        """Test that the probabilities given are legit using law of big numbers."""
        bins = defaultdict(int)
        probs = (0.1, 0.2, 0.3, 0.4)
        categories = ('asdfa', '2', '3', '4')
        categories = OrderedDict(zip(categories, probs))
        dim = Categorical('yolo', categories)
        for _ in range(500):
            sample = dim.sample(seed=seed)[0]
            bins[sample] += 1
        for keys in bins.keys():
            bins[keys] /= float(500)
        for key, value in categories.items():
            assert abs(bins[key] - value) < 0.01

    def test_contains_wrong_shape(self):
        """Check correct category but wrongly shaped array."""
        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 4: 0.4}
        dim = Categorical('yolo', categories, shape=2)

        assert 3 not in dim
        assert ('asdfa', 2) in dim

    def test_repr_too_many_cats(self):
        """Check ellipsis on str/repr of too many categories."""
        categories = tuple(range(10))
        dim = Categorical('yolo', categories, shape=2)

        assert str(dim) == "Categorical(name=yolo, " \
                           "prior={0: 0.10, 1: 0.10, ..., 8: 0.10, 9: 0.10}, " \
                           "shape=(2,), default value=None)"

    def test_bad_probabilities(self):
        """User provided bad probabilities."""
        categories = {'asdfa': 0.05, 2: 0.2, 3: 0.3, 4: 0.4}
        with pytest.raises(ValueError):
            Categorical('yolo', categories, shape=2)

    def test_interval(self):
        """Check that calling `Categorical.interval` raises `RuntimeError`."""
        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 4: 0.4}
        dim = Categorical('yolo', categories, shape=2)

        assert dim.interval() == ('asdfa', 2, 3, 4)

    def test_that_objects_types_are_ok(self):
        """Check that output samples are of the correct type.

        Don't let numpy mess with their automatic type inference.
        """
        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 'lalala': 0.4}
        dim = Categorical('yolo', categories)

        assert '2' not in dim
        assert 2 in dim
        assert 'asdfa' in dim

        dim = Categorical('yolo', categories, shape=(2,))

        assert ['2', 'asdfa'] not in dim
        assert [2, 'asdfa'] in dim

    def test_init_with_default_value_string(self):
        """Make sure the default value is of the correct type"""
        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 'lalala': 0.4}
        dim = Categorical('yolo', categories, default_value="asdfa")

        assert type(dim.default_value) is str

    def test_init_with_default_value_int(self):
        """Make sure the default value is of the correct type"""
        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 'lalala': 0.4}
        dim = Categorical('yolo', categories, default_value=2)

        assert type(dim.default_value) is int

    def test_init_with_wrong_default_value(self):
        """Make sure the default value exists"""
        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 'lalala': 0.4}

        with pytest.raises(ValueError):
            Categorical('yolo', categories, default_value=2.3)

    def test_no_default_value(self):
        """Make sure the default value is None"""
        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 'lalala': 0.4}
        dim = Categorical('yolo', categories)
        assert dim.default_value is None

    def test_cast_list(self):
        """Make sure list are cast to categories and returned as list"""
        categories = {'asdfa': 0.1, 2: 0.2, 3.0: 0.3, 'lalala': 0.4}
        dim = Categorical('yolo', categories)
        assert dim.cast(['asdfa']) == ['asdfa']
        assert dim.cast(['2']) == [2]
        assert dim.cast(['3.0']) == [3.0]

    def test_cast_list_multidim(self):
        """Make sure array are cast to int and returned as array of values"""
        categories = list(range(10))
        categories[0] = 'asdfa'
        categories[2] = 'lalala'
        dim = Categorical('yolo', categories, shape=2)
        sample = ['asdfa', '1']  # np.array(['asdfa', '1'], dtype=np.object)
        assert dim.cast(sample) == ['asdfa', 1]

    def test_cast_array_multidim(self):
        """Make sure array are cast to int and returned as array of values"""
        categories = list(range(10))
        categories[0] = 'asdfa'
        categories[2] = 'lalala'
        dim = Categorical('yolo', categories, shape=2)
        sample = np.array(['asdfa', '1'], dtype=np.object)
        assert np.all(dim.cast(sample) == np.array(['asdfa', 1], dtype=np.object))

    def test_cast_bad_category(self):
        """Make sure array are cast to int and returned as array of values"""
        categories = list(range(10))
        dim = Categorical('yolo', categories, shape=2)
        sample = np.array(['asdfa', '1'], dtype=np.object)
        with pytest.raises(ValueError) as exc:
            dim.cast(sample)
        assert "Invalid category: asdfa" in str(exc.value)


class TestFidelity(object):
    """Test methods of a Fidelity object."""

    def test_simple_instance(self):
        """Test Fidelity.__init__."""
        dim = Fidelity('epoch', 1, 2)

        assert str(dim) == "Fidelity(name=epoch, low=1, high=2, base=2)"
        assert dim.low == 1
        assert dim.high == 2
        assert dim.base == 2
        assert dim.name == 'epoch'
        assert dim.type == 'fidelity'
        assert dim.shape is None

    def test_min_resources(self):
        """Test that an error is raised if min is smaller than 1"""
        with pytest.raises(AttributeError) as exc:
            Fidelity('epoch', 0, 2)
        assert "Minimum resources must be a positive number." == str(exc.value)

    def test_min_max_resources(self):
        """Test that an error is raised if min is larger than max"""
        with pytest.raises(AttributeError) as exc:
            Fidelity('epoch', 3, 2)
        assert "Minimum resources must be smaller than maximum resources." == str(exc.value)

    def test_base(self):
        """Test that an error is raised if base is smaller than 1"""
        with pytest.raises(AttributeError) as exc:
            Fidelity('epoch', 1, 2, 0)
        assert "Base should be greater than or equal to 1" == str(exc.value)

    def test_sampling(self):
        """Make sure Fidelity simply returns `high`"""
        dim = Fidelity('epoch', 1, 2)
        assert dim.sample() == [2]
        dim = Fidelity('epoch', 1, 5)
        assert dim.sample() == [5]
        dim = Fidelity('epoch', 1, 5)
        assert dim.sample(4) == [5] * 4

    def test_default_value(self):
        """Make sure Fidelity simply returns `high`"""
        dim = Fidelity('epoch', 1, 2)
        assert dim.default_value == 2
        dim = Fidelity('epoch', 1, 5)
        assert dim.default_value == 5

    def test_contains(self):
        """Make sure fidelity.__contains__ tests based on (min, max)"""
        dim = Fidelity('epoch', 1, 10)

        assert 0 not in dim
        assert 1 in dim
        assert 5 in dim
        assert 10 in dim
        assert 20 not in dim

    def test_interval(self):
        """Check that interval() is (min, max)."""
        dim = Fidelity('epoch', 1, 10)
        dim.interval() == (1, 10)

    def test_cast(self):
        """Check that error is being raised."""
        dim = Fidelity('epoch', 1, 10)
        with pytest.raises(NotImplementedError):
            dim.cast()


class TestSpace(object):
    """Test methods of a `Space` object."""

    def test_init(self):
        """Instantiate space, must be a dictionary."""
        space = Space()
        assert isinstance(space, dict)

    def test_register_and_contain(self):
        """Register bunch of dimensions, check if points/name are in space."""
        space = Space()

        assert 'yolo' not in space
        assert (('asdfa', 2), 0, 3.5) not in space

        categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 4: 0.4}
        dim = Categorical('yolo', categories, shape=2)
        space.register(dim)
        dim = Integer('yolo2', 'uniform', -3, 6)
        space.register(dim)
        dim = Real('yolo3', 'norm', 0.9)
        space.register(dim)

        assert 'yolo' in space
        assert 'yolo2' in space
        assert 'yolo3' in space

        assert (('asdfa', 2), 0, 3.5) in space
        assert (('asdfa', 2), 7, 3.5) not in space

    def test_bad_contain(self):
        """Checking with no iterables does no good."""
        space = Space()
        with pytest.raises(TypeError):
            5 in space

    def test_sample(self, seed):
        """Check whether sampling works correctly."""
        space = Space()
        probs = (0.1, 0.2, 0.3, 0.4)
        categories = ('asdfa', 2, 3, 4)
        dim1 = Categorical('yolo', OrderedDict(zip(categories, probs)), shape=(2, 2))
        space.register(dim1)
        dim2 = Integer('yolo2', 'uniform', -3, 6)
        space.register(dim2)
        dim3 = Real('yolo3', 'norm', 0.9)
        space.register(dim3)

        point = space.sample(seed=seed)
        test_point = [(dim1.sample()[0],
                       dim2.sample()[0],
                       dim3.sample()[0]), ]
        assert len(point) == len(test_point) == 1
        assert len(point[0]) == len(test_point[0]) == 3
        assert np.all(point[0][0] == test_point[0][0])
        assert point[0][1] == test_point[0][1]
        assert point[0][2] == test_point[0][2]

        points = space.sample(2, seed=seed)
        points1 = dim1.sample(2)
        points2 = dim2.sample(2)
        points3 = dim3.sample(2)
        test_points = [(points1[0], points2[0], points3[0]),
                       (points1[1], points2[1], points3[1])]
        assert len(points) == len(test_points) == 2
        for i in range(2):
            assert len(points[i]) == len(test_points[i]) == 3
            assert np.all(points[i][0] == test_points[i][0])
            assert points[i][1] == test_points[i][1]
            assert points[i][2] == test_points[i][2]

    def test_interval(self):
        """Check whether interval is cool."""
        space = Space()
        probs = (0.1, 0.2, 0.3, 0.4)
        categories = ('asdfa', 2, 3, 4)
        dim = Categorical('yolo', OrderedDict(zip(categories, probs)), shape=2)
        space.register(dim)
        dim = Integer('yolo2', 'uniform', -3, 6)
        space.register(dim)
        dim = Real('yolo3', 'norm', 0.9)
        space.register(dim)

        assert space.interval() == [categories, (-3, 3), (-np.inf, np.inf)]

    def test_cardinality(self):
        """Check whether space capacity is correct"""
        space = Space()
        probs = (0.1, 0.2, 0.3, 0.4)
        categories = ('asdfa', 2, 3, 4)
        dim = Categorical('yolo', OrderedDict(zip(categories, probs)), shape=2)
        space.register(dim)
        dim = Integer('yolo2', 'uniform', -3, 6)
        space.register(dim)
        dim = Fidelity('epoch', 1, 9, 3)
        space.register(dim)

        assert (4 * 2) * 6 * 1 == space.cardinality

        dim = Integer('yolo3', 'uniform', -3, 2, shape=(3, 1))
        space.register(dim)
        assert (4 * 2) * 6 * 1 * (2 * 3 * 1) == space.cardinality

        dim = Real('yolo4', 'norm', 0.9)
        space.register(dim)
        assert np.inf == space.cardinality

    def test_bad_setitem(self):
        """Check exceptions in setting items in Space."""
        space = Space()

        # The name of an integer must be a of `str` type.
        # Integers are reversed for indexing the OrderedDict.
        with pytest.raises(TypeError) as exc:
            space[5] = Integer('yolo', 'uniform', -3, 6)
        assert "string" in str(exc.value)

        # Only object of type `Dimension` are allowed in `Space`.
        with pytest.raises(TypeError) as exc:
            space['ispis'] = 'nope'
        assert "Dimension" in str(exc.value)

        # Cannot register something with the same name.
        space.register(Integer('yolo', 'uniform', -3, 6))
        with pytest.raises(ValueError) as exc:
            space.register(Real('yolo', 'uniform', 0, 6))
        assert "another name" in str(exc.value)

    def test_getitem(self):
        """Test getting dimensions from space."""
        space = Space()
        probs = (0.1, 0.2, 0.3, 0.4)
        categories = ('asdfa', 2, 3, 4)
        dim = Categorical('yolo', OrderedDict(zip(categories, probs)), shape=2)
        space.register(dim)
        dim = Integer('yolo2', 'uniform', -3, 6)
        space.register(dim)
        dim = Real('yolo3', 'norm', 0.9)
        space.register(dim)

        assert space['yolo'].type == 'categorical'
        assert space[0].type == 'categorical'

        with pytest.raises(KeyError):
            space['asdf']

        with pytest.raises(IndexError):
            space[3]

    def test_order(self):
        """Test that the same space built twice will have the same ordering."""
        space1 = Space()
        space1.register(Integer('yolo1', 'uniform', -3, 6, shape=(2,)))
        space1.register(Integer('yolo2', 'uniform', -3, 6, shape=(2,)))
        space1.register(Real('yolo3', 'norm', 0.9))
        space1.register(Categorical('yolo4', ('asdfa', 2)))

        space2 = Space()
        space2.register(Integer('yolo1', 'uniform', -3, 6, shape=(2,)))
        space2.register(Real('yolo3', 'norm', 0.9))
        space2.register(Categorical('yolo4', ('asdfa', 2)))
        space2.register(Integer('yolo2', 'uniform', -3, 6, shape=(2,)))

        assert list(space1) == list(space1.keys())
        assert list(space2) == list(space2.keys())
        assert list(space1.values()) == list(space2.values())
        assert list(space1.items()) == list(space2.items())
        assert list(space1.keys()) == list(space2.keys())
        assert list(space1.values()) == list(space2.values())
        assert list(space1.items()) == list(space2.items())

    def test_repr(self):
        """Test str/repr."""
        space = Space()
        dim = Integer('yolo2', 'uniform', -3, 6, shape=(2,))
        space.register(dim)
        dim = Real('yolo3', 'norm', 0.9)
        space.register(dim)

        assert str(space) == "Space(["\
                             "Integer(name=yolo2, prior={uniform: (-3, 6), {}}, shape=(2,), "\
                             "default value=None),\n"\
                             "       Real(name=yolo3, prior={norm: (0.9,), {}}, shape=(), "\
                             "default value=None)])"

    def test_configuration(self):
        """Test that configuration contains all dimensions."""
        space = Space()
        space.register(Integer('yolo1', 'uniform', -3, 6, shape=(2,)))
        space.register(Integer('yolo2', 'uniform', -3, 6, shape=(2,)))
        space.register(Real('yolo3', 'norm', 0.9))
        space.register(Categorical('yolo4', ('asdfa', 2)))

        assert space.configuration == {
            'yolo1': 'uniform(-3, 3, shape=(2,), discrete=True)',
            'yolo2': 'uniform(-3, 3, shape=(2,), discrete=True)',
            'yolo3': 'norm(0.9)',
            'yolo4': 'choices([\'asdfa\', 2])'}

    def test_precision(self):
        """Test that precision is correctly handled."""
        space = Space()
        space.register(Real('yolo1', 'norm', 0.9, precision=6))
        space.register(Real('yolo2', 'norm', 0.9, precision=None))
        space.register(Real('yolo5', 'norm', 0.9))

        assert space['yolo1'].precision == 6
        assert space['yolo2'].precision is None
        assert space['yolo5'].precision == 4

        with pytest.raises(TypeError):
            space.register(Real('yolo3', 'norm', 0.9, precision=-12))

        with pytest.raises(TypeError):
            space.register(Real('yolo4', 'norm', 0.9, precision=0.6))



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io` -- Input/output functions for interaction with user's script
=================================================================================

.. module:: io
   :platform: Unix
   :synopsis: Parsers and formatters to be used with user's input files or cli
      arguments and expected outputs.
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Common fixtures and utils for unittests and functional tests."""
import copy
import os

from pymongo import MongoClient
import pytest
import yaml

from orion.algo.base import (BaseAlgorithm, OptimizationAlgorithm)
import orion.core.cli
from orion.core.io.database import Database
import orion.core.io.experiment_builder as experiment_builder
import orion.core.utils.backward as backward
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage


class DumbAlgo(BaseAlgorithm):
    """Stab class for `BaseAlgorithm`."""

    def __init__(self, space, value=5,
                 scoring=0, judgement=None,
                 suspend=False, done=False, **nested_algo):
        """Configure returns, allow for variable variables."""
        self._times_called_suspend = 0
        self._times_called_is_done = 0
        self._num = None
        self._points = None
        self._results = None
        self._score_point = None
        self._judge_point = None
        self._measurements = None
        super(DumbAlgo, self).__init__(space, value=value,
                                       scoring=scoring, judgement=judgement,
                                       suspend=suspend,
                                       done=done,
                                       **nested_algo)

    def suggest(self, num=1):
        """Suggest based on `value`."""
        self._num = num
        return [self.value] * num

    def observe(self, points, results):
        """Log inputs."""
        self._points = points
        self._results = results

    def score(self, point):
        """Log and return stab."""
        self._score_point = point
        return self.scoring

    def judge(self, point, measurements):
        """Log and return stab."""
        self._judge_point = point
        self._measurements = measurements
        return self.judgement

    @property
    def should_suspend(self):
        """Cound how many times it has been called and return `suspend`."""
        self._times_called_suspend += 1
        return self.suspend

    @property
    def is_done(self):
        """Cound how many times it has been called and return `done`."""
        self._times_called_is_done += 1
        return self.done


# Hack it into being discoverable
OptimizationAlgorithm.types.append(DumbAlgo)
OptimizationAlgorithm.typenames.append(DumbAlgo.__name__.lower())


@pytest.fixture(scope='session')
def dumbalgo():
    """Return stab algorithm class."""
    return DumbAlgo


@pytest.fixture()
def exp_config():
    """Load an example database."""
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
              'experiment.yaml')) as f:
        exp_config = list(yaml.safe_load_all(f))

    for config in exp_config[0]:
        backward.populate_space(config)

    return exp_config


@pytest.fixture(scope='session')
def database():
    """Return Mongo database object to test with example entries."""
    client = MongoClient(username='user', password='pass', authSource='orion_test')
    database = client.orion_test
    yield database
    client.close()


@pytest.fixture()
def clean_db(database, db_instance):
    """Clean insert example experiment entries to collections."""
    database.experiments.drop()
    database.lying_trials.drop()
    database.trials.drop()
    database.workers.drop()
    database.resources.drop()


@pytest.fixture()
def db_instance(null_db_instances):
    """Create and save a singleton database instance."""
    try:
        db = Database(of_type='MongoDB', name='orion_test',
                      username='user', password='pass')
    except ValueError:
        db = Database()

    return db


@pytest.fixture
def only_experiments_db(clean_db, database, exp_config):
    """Clean the database and insert only experiments."""
    database.experiments.insert_many(exp_config[0])


def ensure_deterministic_id(name, db_instance, version=1, update=None):
    """Change the id of experiment to its name."""
    experiment = db_instance.read('experiments', {'name': name, 'version': version})[0]
    db_instance.remove('experiments', {'_id': experiment['_id']})
    _id = name + "_" + str(version)
    experiment['_id'] = _id

    if experiment['refers']['parent_id'] is None:
        experiment['refers']['root_id'] = _id

    if update is not None:
        experiment.update(update)

    db_instance.write('experiments', experiment)


# Experiments combinations fixtures
@pytest.fixture
def one_experiment(monkeypatch, db_instance):
    """Create an experiment without trials."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    name = 'test_single_exp'
    orion.core.cli.main(['hunt', '--init-only', '-n', name,
                         './black_box.py', '--x~uniform(0,1)'])
    ensure_deterministic_id(name, db_instance)
    return get_storage().fetch_experiments({'name': name})[0]


@pytest.fixture
def one_experiment_changed_vcs(one_experiment):
    """Create an experiment without trials."""
    experiment = experiment_builder.build(name=one_experiment['name'])

    experiment.metadata['VCS'] = {
        'type': 'git', 'is_dirty': False, 'HEAD_sha': 'new', 'active_branch': 'master',
        'diff_sha': None}

    get_storage().update_experiment(experiment, metadata=experiment.metadata)


@pytest.fixture
def one_experiment_no_version(monkeypatch, one_experiment):
    """Create an experiment without trials."""
    one_experiment['name'] = one_experiment['name'] + '-no-version'
    one_experiment.pop('version')

    def fetch_without_version(query, selection=None):
        if query.get('name') == one_experiment['name'] or query == {}:
            return [copy.deepcopy(one_experiment)]

        return []

    monkeypatch.setattr(get_storage(), 'fetch_experiments', fetch_without_version)

    return one_experiment


@pytest.fixture
def with_experiment_using_python_api(monkeypatch, one_experiment):
    """Create an experiment without trials."""
    experiment = experiment_builder.build(name='from-python-api', space={'x': 'uniform(0, 10)'})

    return experiment


@pytest.fixture
def with_experiment_missing_conf_file(monkeypatch, one_experiment):
    """Create an experiment without trials."""
    exp = experiment_builder.build(name='test_single_exp', version=1)
    conf_file = 'idontexist.yaml'
    exp.metadata['user_config'] = conf_file
    exp.metadata['user_args'] += ['--config', conf_file]
    Database().write('experiments', exp.configuration, query={'_id': exp.id})

    return exp


@pytest.fixture
def broken_refers(one_experiment, db_instance):
    """Create an experiment with broken refers."""
    ensure_deterministic_id('test_single_exp', db_instance, update=dict(refers={'oups': 'broken'}))


@pytest.fixture
def single_without_success(one_experiment):
    """Create an experiment without a succesful trial."""
    statuses = list(Trial.allowed_stati)
    statuses.remove('completed')

    exp = experiment_builder.build(name='test_single_exp')
    x = {'name': '/x', 'type': 'real'}

    x_value = 0
    for status in statuses:
        x['value'] = x_value
        trial = Trial(experiment=exp.id, params=[x], status=status)
        x_value += 1
        Database().write('trials', trial.to_dict())


@pytest.fixture
def single_with_trials(single_without_success):
    """Create an experiment with all types of trials."""
    exp = experiment_builder.build(name='test_single_exp')

    x = {'name': '/x', 'type': 'real', 'value': 100}
    results = {"name": "obj", "type": "objective", "value": 0}
    trial = Trial(experiment=exp.id, params=[x], status='completed', results=[results])
    Database().write('trials', trial.to_dict())


@pytest.fixture
def two_experiments(monkeypatch, db_instance):
    """Create an experiment and its child."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_double_exp',
                         './black_box.py', '--x~uniform(0,1)'])
    ensure_deterministic_id('test_double_exp', db_instance)

    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_double_exp',
                         '--branch-to', 'test_double_exp_child', './black_box.py',
                         '--x~+uniform(0,1,default_value=0)', '--y~+uniform(0,1,default_value=0)'])
    ensure_deterministic_id('test_double_exp_child', db_instance)


@pytest.fixture
def family_with_trials(two_experiments):
    """Create two related experiments with all types of trials."""
    exp = experiment_builder.build(name='test_double_exp')
    exp2 = experiment_builder.build(name='test_double_exp_child')
    x = {'name': '/x', 'type': 'real'}
    y = {'name': '/y', 'type': 'real'}

    x_value = 0
    for status in Trial.allowed_stati:
        x['value'] = x_value
        y['value'] = x_value
        trial = Trial(experiment=exp.id, params=[x], status=status)
        x['value'] = x_value
        trial2 = Trial(experiment=exp2.id, params=[x, y], status=status)
        x_value += 1
        Database().write('trials', trial.to_dict())
        Database().write('trials', trial2.to_dict())


@pytest.fixture
def unrelated_with_trials(family_with_trials, single_with_trials):
    """Create two unrelated experiments with all types of trials."""
    exp = experiment_builder.build(name='test_double_exp_child')

    Database().remove('trials', {'experiment': exp.id})
    Database().remove('experiments', {'_id': exp.id})


@pytest.fixture
def three_experiments(two_experiments, one_experiment):
    """Create a single experiment and an experiment and its child."""
    pass


@pytest.fixture
def three_experiments_with_trials(family_with_trials, single_with_trials):
    """Create three experiments, two unrelated, with all types of trials."""
    pass


@pytest.fixture
def three_experiments_family(two_experiments, db_instance):
    """Create three experiments, one of which is the parent of the other two."""
    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_double_exp',
                         '--branch-to', 'test_double_exp_child2', './black_box.py',
                         '--x~+uniform(0,1,default_value=0)', '--z~+uniform(0,1,default_value=0)'])
    ensure_deterministic_id('test_double_exp_child2', db_instance)


@pytest.fixture
def three_family_with_trials(three_experiments_family, family_with_trials):
    """Create three experiments, all related, two direct children, with all types of trials."""
    exp = experiment_builder.build(name='test_double_exp_child2')
    x = {'name': '/x', 'type': 'real'}
    z = {'name': '/z', 'type': 'real'}

    x_value = 0
    for status in Trial.allowed_stati:
        x['value'] = x_value
        z['value'] = x_value * 100
        trial = Trial(experiment=exp.id, params=[x, z], status=status)
        x_value += 1
        Database().write('trials', trial.to_dict())


@pytest.fixture
def three_experiments_family_branch(two_experiments, db_instance):
    """Create three experiments, each parent of the following one."""
    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_double_exp_child',
                         '--branch-to', 'test_double_exp_grand_child', './black_box.py',
                         '--x~+uniform(0,1,default_value=0)', '--y~uniform(0,1,default_value=0)',
                         '--z~+uniform(0,1,default_value=0)'])
    ensure_deterministic_id('test_double_exp_grand_child', db_instance)


@pytest.fixture
def three_family_branch_with_trials(three_experiments_family_branch, family_with_trials):
    """Create three experiments, all related, one child and one grandchild,
    with all types of trials.

    """
    exp = experiment_builder.build(name='test_double_exp_grand_child')
    x = {'name': '/x', 'type': 'real'}
    y = {'name': '/y', 'type': 'real'}
    z = {'name': '/z', 'type': 'real'}

    x_value = 0
    for status in Trial.allowed_stati:
        x['value'] = x_value
        y['value'] = x_value * 10
        z['value'] = x_value * 100
        trial = Trial(experiment=exp.id, params=[x, y, z], status=status)
        x_value += 1
        Database().write('trials', trial.to_dict())


@pytest.fixture
def two_experiments_same_name(one_experiment, db_instance):
    """Create two experiments with the same name but different versions."""
    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_single_exp',
                         './black_box.py', '--x~uniform(0,1)', '--y~+normal(0,1)'])
    ensure_deterministic_id('test_single_exp', db_instance, version=2)


@pytest.fixture
def three_experiments_family_same_name(two_experiments_same_name, db_instance):
    """Create three experiments, two of them with the same name but different versions and one
    with a child.
    """
    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_single_exp', '-v', '1', '-b',
                         'test_single_exp_child', './black_box.py', '--x~uniform(0,1)',
                         '--y~+normal(0,1)'])
    ensure_deterministic_id('test_single_exp_child', db_instance)


@pytest.fixture
def three_experiments_branch_same_name(two_experiments_same_name, db_instance):
    """Create three experiments, two of them with the same name but different versions and last one
    with a child.
    """
    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_single_exp', '-b',
                         'test_single_exp_child', './black_box.py', '--x~uniform(0,1)',
                         '--y~normal(0,1)', '--z~+normal(0,1)'])
    ensure_deterministic_id('test_single_exp_child', db_instance)


@pytest.fixture
def three_experiments_same_name(two_experiments_same_name, db_instance):
    """Create three experiments with the same name but different versions."""
    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_single_exp',
                         './black_box.py', '--x~uniform(0,1)', '--y~normal(0,1)',
                         '--z~+normal(0,1)'])
    ensure_deterministic_id('test_single_exp', db_instance, version=3)


@pytest.fixture
def three_experiments_same_name_with_trials(two_experiments_same_name, db_instance):
    """Create three experiments with the same name but different versions."""
    orion.core.cli.main(['hunt', '--init-only', '-n', 'test_single_exp',
                         './black_box.py', '--x~uniform(0,1)', '--y~normal(0,1)',
                         '--z~+normal(0,1)'])
    ensure_deterministic_id('test_single_exp', db_instance, version=3)

    exp = experiment_builder.build(name='test_single_exp', version=1)
    exp2 = experiment_builder.build(name='test_single_exp', version=2)
    exp3 = experiment_builder.build(name='test_single_exp', version=3)

    x = {'name': '/x', 'type': 'real'}
    y = {'name': '/y', 'type': 'real'}
    z = {'name': '/z', 'type': 'real'}
    x_value = 0
    for status in Trial.allowed_stati:
        x['value'] = x_value
        y['value'] = x_value * 10
        z['value'] = x_value * 100
        trial = Trial(experiment=exp.id, params=[x], status=status)
        trial2 = Trial(experiment=exp2.id, params=[x, y], status=status)
        trial3 = Trial(experiment=exp3.id, params=[x, y, z], status=status)
        Database().write('trials', trial.to_dict())
        Database().write('trials', trial2.to_dict())
        Database().write('trials', trial3.to_dict())
        x_value += 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform functional tests for the parsing of the basic arguments groups."""
import argparse
import os

import pytest

import orion.core.cli.base as cli


def _create_parser(need_subparser=True):
    parser = argparse.ArgumentParser()

    if need_subparser:
        subparsers = parser.add_subparsers()
        return parser, subparsers

    return parser


@pytest.mark.usefixtures("clean_db")
def test_common_group_arguments(database, monkeypatch):
    """Check the parsing of the common group"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    parser, subparsers = _create_parser()
    args_list = ["-n", "test", "--config", "./orion_config_random.yaml"]

    cli.get_basic_args_group(parser)
    args = vars(parser.parse_args(args_list))
    assert args['name'] == 'test'
    assert args['config'].name == "./orion_config_random.yaml"


@pytest.mark.usefixtures("clean_db")
def test_user_group_arguments(database, monkeypatch):
    """Test the parsing of the user group"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    parser = _create_parser(False)
    args_list = ["./black_box.py", "-x~normal(50,50)"]

    cli.get_user_args_group(parser)
    args = vars(parser.parse_args(args_list))
    assert len(args['user_args']) == 2
    assert args['user_args'] == ['./black_box.py', '-x~normal(50,50)']


@pytest.mark.usefixtures("clean_db")
def test_common_and_user_group_arguments(database, monkeypatch):
    """Test the parsing of the command and user groups"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    parser = _create_parser(False)
    args_list = ["-n", "test", "-c", "./orion_config_random.yaml",
                 "./black_box.py", "-x~normal(50,50)"]

    cli.get_basic_args_group(parser)
    cli.get_user_args_group(parser)
    args = vars(parser.parse_args(args_list))
    assert args['name'] == 'test'
    assert args['config'].name == './orion_config_random.yaml'
    assert len(args['user_args']) == 2
    assert args['user_args'] == ['./black_box.py', '-x~normal(50,50)']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.cmdline_parser`."""
from collections import OrderedDict
import os

import pytest

from orion.core.io.cmdline_parser import CmdlineParser


@pytest.fixture
def basic_config():
    """Return a simple configuration"""
    config = OrderedDict()

    config['_pos_0'] = 'python'
    config['_pos_1'] = 'script.py'
    config['_pos_2'] = 'some'
    config['_pos_3'] = 'pos'
    config['_pos_4'] = 'args'
    config['with'] = 'args'
    config['and'] = ['multiple', 'args']
    config['plus'] = True
    config['booleans'] = True
    config['equal'] = 'value'

    return config


@pytest.fixture
def basic_keys():
    """Return keys of a simple configuration"""
    keys = OrderedDict()

    keys['_pos_0'] = '_pos_0'
    keys['_pos_1'] = '_pos_1'
    keys['_pos_2'] = '_pos_2'
    keys['_pos_3'] = '_pos_3'
    keys['_pos_4'] = '_pos_4'
    keys['with'] = '--with'
    keys['and'] = '--and'
    keys['plus'] = '--plus'
    keys['booleans'] = '--booleans'
    keys['equal'] = '--equal'

    return keys


@pytest.fixture
def to_format():
    """Return a commandline to format"""
    return "python 1 --arg value --args value1 value2 --boolean"


def test_key_to_arg():
    """Test the key to arg function"""
    cmdline_parser = CmdlineParser()
    assert cmdline_parser._key_to_arg("c") == "-c"
    assert cmdline_parser._key_to_arg("test") == "--test"
    assert cmdline_parser._key_to_arg("test-test") == "--test-test"


def test_parse_paths(monkeypatch):
    """Test the parse_paths function"""
    monkeypatch.chdir(os.path.dirname(__file__))
    cmdline_parser = CmdlineParser()
    assert cmdline_parser._parse_paths(__file__) == os.path.abspath(__file__)

    values = ['test_resolve_config.py', 'test', 'nada', __file__]
    parsed_values = cmdline_parser._parse_paths(values)
    assert parsed_values[0] == os.path.abspath('test_resolve_config.py')
    assert parsed_values[1] == 'test'
    assert parsed_values[2] == 'nada'
    assert parsed_values[3] == os.path.abspath(__file__)


def test_parse_arguments(basic_config, basic_keys):
    """Test the parsing of the commandline arguments"""
    cmdline_parser = CmdlineParser()
    keys, arguments = cmdline_parser._parse_arguments(
        "python script.py some pos args --with args --and multiple args "
        "--plus --booleans --equal=value".split(" "))

    assert arguments == basic_config
    assert keys == basic_keys


def test_parse_arguments_template():
    """Test the creation of the template for arguments"""
    cmdline_parser = CmdlineParser()

    cmdline_parser.parse(
        "python script.py some pos args "
        "--with args --and multiple args --plus --booleans --equal=value".split(" "))

    assert (
        cmdline_parser.template ==
        ['{_pos_0}', '{_pos_1}', '{_pos_2}', '{_pos_3}', '{_pos_4}', '--with', '{with}', '--and',
         '{and[0]}', '{and[1]}', '--plus', '--booleans', '--equal', '{equal}'])


def test_format(to_format):
    """Test that the format method assigns the correc values"""
    cmdline_parser = CmdlineParser()

    cmdline_parser.parse(to_format.split(' '))

    keys, arguments = cmdline_parser._parse_arguments(to_format.split(' '))

    formatted = cmdline_parser.format(arguments)

    assert formatted == to_format.split(' ')


def test_parse_arguments_bad_command():
    """Test the fail cases of parsing"""
    cmdline_parser = CmdlineParser()

    with pytest.raises(ValueError) as exc_info:
        cmdline_parser.parse(
            "python script.py some pos args "
            "--with args --and multiple args --plus --booleans "
            "--and dummy.yaml".split(" "))

    assert "Conflict: two arguments have the same name: and" in str(exc_info.value)


def test_has_already_been_parsed():
    """Test the template from the branching"""
    cmdline_parser = CmdlineParser()

    command = "python script.py some pos args " \
              "--with args --and multiple args --plus --booleans "

    cmdline_parser.parse(command.split(' '))
    with pytest.raises(RuntimeError) as exc_info:
        cmdline_parser.parse(command.split(' '))

    assert "already" in str(exc_info.value)


def test_get_state_dict():
    """Test getting state dict."""
    cmdline_parser = CmdlineParser()

    cmdline_parser.parse(
        "python script.py some pos args "
        "--with args --and multiple args --plus --booleans --equal=value".split(" "))

    assert cmdline_parser.get_state_dict() == {
        'keys': list(map(list, cmdline_parser.keys.items())),
        'arguments': list(map(list, cmdline_parser.arguments.items())),
        'template': cmdline_parser.template}


def test_set_state_dict():
    """Test that set_state_dict sets state properly to generate new cmdline."""
    cmdline_parser = CmdlineParser()

    cmdline_parser.set_state_dict({
        'arguments': {},
        'template': ['{_pos_0}', '{_pos_1}', '--with', '{with}', '--plus', '--booleans']})

    assert cmdline_parser.format({'_pos_0': 'voici', '_pos_1': 'voila', 'with': 'classe'}) == [
        'voici', 'voila', '--with', 'classe', '--plus', '--booleans']


def test_parse_not_enough_dashes():
    """Test that arguments with many chars but one dash are supported even if it is not standard"""
    cmdline_parser = CmdlineParser()
    keys, arguments = cmdline_parser._parse_arguments(
        "pos -not-enough dashes "
        "--enough dashes -o my".split(" "))

    assert arguments == OrderedDict((
        ('_pos_0', 'pos'),
        ('not-enough', 'dashes'),
        ('enough', 'dashes'),
        ('o', 'my')))

    assert keys == OrderedDict((
        ('_pos_0', '_pos_0'),
        ('not-enough', '-not-enough'),
        ('enough', '--enough'),
        ('o', '-o')))


def test_parse_fugly_underscores():
    """Test that underscores are kept as such no matter how fugly this is"""
    cmdline_parser = CmdlineParser()
    keys, arguments = cmdline_parser._parse_arguments(
        "pos -my_poor eyes --are_bleeding because --of-these underscores".split(" "))

    assert arguments == OrderedDict((
        ('_pos_0', 'pos'),
        ('my_poor', 'eyes'),
        ('are_bleeding', 'because'),
        ('of-these', 'underscores')))

    assert keys == OrderedDict((
        ('_pos_0', '_pos_0'),
        ('my_poor', '-my_poor'),
        ('are_bleeding', '--are_bleeding'),
        ('of-these', '--of-these')))



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli` -- Base class and function utilities for cli
==================================================================

.. module:: cli
   :platform: Unix
   :synopsis: Orion main parser class and helper functions to parse command-line options

"""
import argparse
import logging
import sys
import textwrap

import orion
from orion.core.io.database import DatabaseError
from orion.core.utils.exceptions import (
    BranchingEvent, InexecutableUserScript, MissingResultFile, NoConfigurationError, NoNameError)


CLI_DOC_HEADER = "Oríon CLI for asynchronous distributed optimization"


class OrionArgsParser:
    """Parser object handling the upper-level parsing of Oríon's arguments."""

    def __init__(self, description=CLI_DOC_HEADER):
        """Create the pre-command arguments"""
        self.description = description

        self.parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent(description))

        self.parser.add_argument(
            '-V', '--version',
            action='version', version='orion ' + orion.core.__version__)

        self.parser.add_argument(
            '-v', '--verbose',
            action='count', default=0,
            help="logging levels of information about the process (-v: INFO. -vv: DEBUG)")

        self.parser.add_argument(
            '-d', '--debug', action='store_true',
            help="Use debugging mode with EphemeralDB.")

        self.subparsers = self.parser.add_subparsers(dest='command')

    def get_subparsers(self):
        """Return the subparser object for this parser."""
        return self.subparsers

    def parse(self, argv):
        """Call argparse and generate a dictionary of arguments' value"""
        args = vars(self.parser.parse_args(argv))

        verbose = args.pop('verbose', 0)
        levels = {0: logging.WARNING,
                  1: logging.INFO,
                  2: logging.DEBUG}
        logging.basicConfig(level=levels.get(verbose, logging.DEBUG))

        if args['command'] is None:
            self.parser.parse_args(['--help'])

        function = args.pop('func', None)
        empty_command = (argv[-1] if argv else sys.argv[-1]) == args['command']
        if function is None or (empty_command and args.pop('help_empty', False)):
            self.parser.parse_args([args['command'], '--help'])

        return args, function

    def execute(self, argv):
        """Execute main function of the subparser"""
        try:
            args, function = self.parse(argv)
            returncode = function(args)
        except (NoConfigurationError, NoNameError, DatabaseError, MissingResultFile,
                BranchingEvent, InexecutableUserScript) as e:
            print('Error:', e, file=sys.stderr)

            if args.get('verbose', 0) >= 2:
                raise e

            return 1

        except KeyboardInterrupt:
            print('Orion is interrupted.')
            return 130

        return 0 if returncode is None else returncode


def get_basic_args_group(
        parser,
        group_name="Oríon arguments",
        group_help="These arguments determine orion's behaviour"):
    """Return the basic arguments for any command."""
    basic_args_group = parser.add_argument_group(
        group_name, description=group_help)

    basic_args_group.add_argument(
        '-n', '--name',
        type=str, metavar='stringID',
        help="experiment's unique name; "
             "(default: None - specified either here or in a config)")

    basic_args_group.add_argument(
        '-u', '--user',
        type=str,
        help="user associated to experiment's unique name; "
             "(default: $USER - can be overriden either here or in a config)")

    basic_args_group.add_argument(
        '-v', '--version', type=int,
        help="specific version of experiment to fetch; "
             "(default: None - latest experiment.)")

    basic_args_group.add_argument('-c', '--config', type=argparse.FileType('r'),
                                  metavar='path-to-config', help="user provided "
                                  "orion configuration file")

    return basic_args_group


def get_user_args_group(parser):
    """
    Return the user group arguments for any command.
    User group arguments are composed of the user script and the user args
    """
    usergroup = parser.add_argument_group(
        "User script related arguments",
        description="These arguments determine user's script behaviour "
                    "and they can serve as orion's parameter declaration.")

    usergroup.add_argument(
        'user_args', nargs=argparse.REMAINDER, metavar='...',
        help="Command line of user script. A configuration "
             "file intended to be used with 'userscript' must be given as a path "
             "in the **first positional** argument OR using `--config=<path>` "
             "keyword argument.")

    return usergroup



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform functional tests for db set."""
import orion.core.cli
import orion.core.io.experiment_builder as experiment_builder
from orion.storage.base import get_storage


def execute(command, assert_code=0):
    """Execute orion command and return returncode"""
    returncode = orion.core.cli.main(command.split(' '))
    assert returncode == assert_code


def test_no_exp(clean_db, capsys):
    """Test that set non-existing exp exits gracefully"""
    execute('db set i-dont-exist whatever=1 idontcare=2', assert_code=1)

    captured = capsys.readouterr()

    assert captured.err.startswith("Error: No experiment with given name 'i-dont-exist'")


def test_confirm_name(monkeypatch, clean_db, single_with_trials):
    """Test name must be confirmed for update"""
    def incorrect_name(*args):
        return 'oops'

    monkeypatch.setattr('builtins.input', incorrect_name)

    execute('db set test_single_exp status=broken status=interrupted', assert_code=1)

    def correct_name(*args):
        return 'test_single_exp'

    monkeypatch.setattr('builtins.input', correct_name)

    assert len(get_storage()._fetch_trials({'status': 'broken'})) > 0
    execute('db set test_single_exp status=broken status=interrupted')
    assert len(get_storage()._fetch_trials({'status': 'broken'})) == 0


def test_invalid_query(clean_db, single_with_trials, capsys):
    """Test error message when query is invalid"""
    execute('db set -f test_single_exp whatever=1 idontcare=2', assert_code=1)

    captured = capsys.readouterr()

    assert captured.err.startswith("Error: Invalid query attribute `whatever`.")


def test_invalid_update(clean_db, single_with_trials, capsys):
    """Test error message when update attribute is invalid"""
    execute('db set -f test_single_exp status=new yoopidoo=2', assert_code=1)

    captured = capsys.readouterr()

    assert captured.err.startswith("Error: Invalid update attribute `yoopidoo`.")


def test_update_trial(clean_db, single_with_trials, capsys):
    """Test that trial is updated properly"""
    trials = get_storage()._fetch_trials({})
    assert sum(trial.status == 'broken' for trial in trials) > 0
    trials = dict(zip((trial.id for trial in trials), trials))
    execute('db set -f test_single_exp status=broken status=interrupted')
    for trial in get_storage()._fetch_trials({}):
        if trials[trial.id].status == 'broken':
            assert trial.status == 'interrupted', 'status not changed properly'
        else:
            assert trials[trial.id].status == trial.status, 'status should not have been changed'

    captured = capsys.readouterr()

    assert captured.out.endswith("1 trials modified\n")


def test_update_trial_with_id(clean_db, single_with_trials, capsys):
    """Test that trial is updated properly when querying with the id"""
    trials = get_storage()._fetch_trials({})
    trials = dict(zip((trial.id for trial in trials), trials))
    trial = get_storage()._fetch_trials({'status': 'broken'})[0]
    execute(f'db set -f test_single_exp id={trial.id} status=interrupted')
    for new_trial in get_storage()._fetch_trials({}):
        if new_trial.id == trial.id:
            assert new_trial.status == 'interrupted', 'status not changed properly'
        else:
            status_unchanged = trials[new_trial.id].status == new_trial.status
            assert status_unchanged, 'status should not have been changed'

    captured = capsys.readouterr()

    assert captured.out.endswith("1 trials modified\n")


def test_update_no_match_query(clean_db, single_with_trials, capsys):
    """Test that no trials are updated when there is no match"""
    trials = get_storage()._fetch_trials({})
    trials = dict(zip((trial.id for trial in trials), trials))
    execute('db set -f test_single_exp status=invalid status=interrupted')
    for trial in get_storage()._fetch_trials({}):
        assert trials[trial.id].status == trial.status, 'status should not have been changed'

    captured = capsys.readouterr()

    assert captured.out.endswith("0 trials modified\n")


def test_update_child_only(clean_db, three_experiments_same_name_with_trials, capsys):
    """Test trials of the parent experiment are not updated"""
    exp1 = experiment_builder.build_view(name='test_single_exp', version=1)
    exp2 = experiment_builder.build_view(name='test_single_exp', version=2)
    exp3 = experiment_builder.build_view(name='test_single_exp', version=3)
    execute('db set -f test_single_exp --version 2 status=broken status=interrupted')
    assert len(exp1.fetch_trials_by_status('broken')) > 0
    assert len(exp2.fetch_trials_by_status('broken')) == 0
    assert len(exp3.fetch_trials_by_status('broken')) == 0

    captured = capsys.readouterr()

    assert captured.out.endswith("2 trials modified\n")


def test_update_default_leaf(clean_db, three_experiments_same_name_with_trials, capsys):
    """Test trials of the parent experiment are not updated"""
    exp1 = experiment_builder.build_view(name='test_single_exp', version=1)
    exp2 = experiment_builder.build_view(name='test_single_exp', version=2)
    exp3 = experiment_builder.build_view(name='test_single_exp', version=3)
    execute('db set -f test_single_exp status=broken status=interrupted')
    assert len(exp1.fetch_trials_by_status('broken')) > 0
    assert len(exp2.fetch_trials_by_status('broken')) > 0
    assert len(exp3.fetch_trials_by_status('broken')) == 0

    captured = capsys.readouterr()

    assert captured.out.endswith("1 trials modified\n")


def test_no_update_id_from_parent(clean_db, three_experiments_same_name_with_trials, capsys):
    """Test trials of the parent experiment are not updated"""
    exp1 = experiment_builder.build_view(name='test_single_exp', version=1)
    exp2 = experiment_builder.build_view(name='test_single_exp', version=2)
    exp3 = experiment_builder.build_view(name='test_single_exp', version=3)
    trial = exp1.fetch_trials_by_status('broken')[0]
    execute(f'db set -f test_single_exp id={trial.id} status=shouldnothappen')
    assert len(exp1.fetch_trials_by_status('shouldnothappen')) == 0
    assert len(exp2.fetch_trials_by_status('shouldnothappen')) == 0
    assert len(exp3.fetch_trials_by_status('shouldnothappen')) == 0

    captured = capsys.readouterr()

    assert captured.out.endswith("0 trials modified\n")



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.worker.strategy` -- register objectives for incomplete trials
========================================================================

.. module:: strategy
   :platform: Unix
   :synopsis: Strategies to register objectives for incomplete trials.

"""
from abc import (ABCMeta, abstractmethod)
import logging

from orion.core.utils import Factory
from orion.core.worker.trial import Trial

log = logging.getLogger(__name__)


CORRUPTED_DB_WARNING = """\
Trial `%s` has an objective but status is not completed.
This is likely due to a corrupted database, possibly because of
database timeouts. Try setting manually status to `completed`.
You can find documention to do this at
https://orion.readthedocs.io/en/stable/user/storage.html#storage-backend.

If you encounter this issue often, please consider reporting it to
https://github.com/Epistimio/orion/issues."""


def get_objective(trial):
    """Get the value for the objective, if it exists, for this trial

    :return: Float or None
        The value of the objective, or None if it doesn't exist
    """
    objectives = [result.value for result in trial.results
                  if result.type == 'objective']

    if not objectives:
        objective = None
    elif len(objectives) == 1:
        objective = objectives[0]
    elif len(objectives) > 1:
        raise RuntimeError("Trial {} has {} objectives".format(trial.id, len(objectives)))

    return objective


class BaseParallelStrategy(object, metaclass=ABCMeta):
    """Strategy to give intermediate results for incomplete trials"""

    @abstractmethod
    def observe(self, points, results):
        """Observe completed trials

        .. seealso:: `orion.algo.base.BaseAlgorithm.observe` method

        :param points: list of tuples of array-likes
           Points from a `orion.algo.space.Space`.
           Evaluated problem parameters by a consumer.
        :param results : list of dicts
           Contains the result of an evaluation; partial information about the
           black-box function at each point in `params`.
        """
        # NOTE: In future points and results will be converted to trials for coherence with
        # `Strategy.lie()` as well as for coherence with `Algorithm.observe` which will also be
        # converted to expect trials instead of lists and dictionaries.
        pass

    # pylint: disable=no-self-use
    def lie(self, trial):
        """Construct a fake result for an incomplete trial

        Parameters
        ----------
        trial: `orion.core.worker.trial.Trial`
            A trial object which is not supposed to be completed.

        Returns
        -------
        ``orion.core.worker.trial.Trial.Result``
            The fake objective result corresponding to the trial given.

        Notes
        -----
        If the trial has an objective even if not completed, a warning is printed to user
        with a pointer to documentation to resolve the database corruption. The result returned is
        the corresponding objective instead of the lie.

        """
        objective = get_objective(trial)
        if objective:
            log.warning(CORRUPTED_DB_WARNING, trial.id)
            return Trial.Result(name='lie', type='lie', value=objective)

        return None

    @property
    def configuration(self):
        """Provide the configuration of the strategy as a dictionary."""
        # TODO(mnoukhov): change to dict {of_type: __name__} ?
        return self.__class__.__name__


class NoParallelStrategy(BaseParallelStrategy):
    """No parallel strategy"""

    def observe(self, points, results):
        """See BaseParallelStrategy.observe"""
        pass

    def lie(self, trial):
        """See BaseParallelStrategy.lie"""
        result = super(NoParallelStrategy, self).lie(trial)
        if result:
            return result

        return None


class MaxParallelStrategy(BaseParallelStrategy):
    """Parallel strategy that uses the max of completed objectives"""

    def __init__(self, default_result=float('inf')):
        """Initialize the maximum result used to lie"""
        self.max_result = default_result

    def observe(self, points, results):
        """See BaseParallelStrategy.observe"""
        super(MaxParallelStrategy, self).observe(points, results)
        self.max_result = max(result['objective'] for result in results
                              if result['objective'] is not None)

    def lie(self, trial):
        """See BaseParallelStrategy.lie"""
        result = super(MaxParallelStrategy, self).lie(trial)
        if result:
            return result

        return Trial.Result(name='lie', type='lie', value=self.max_result)


class MeanParallelStrategy(BaseParallelStrategy):
    """Parallel strategy that uses the mean of completed objectives"""

    def __init__(self, default_result=float('inf')):
        """Initialize the mean result used to lie"""
        self.mean_result = default_result

    def observe(self, points, results):
        """See BaseParallelStrategy.observe"""
        super(MeanParallelStrategy, self).observe(points, results)
        objective_values = [result['objective'] for result in results
                            if result['objective'] is not None]
        self.mean_result = sum(value for value in objective_values) / float(len(objective_values))

    def lie(self, trial):
        """See BaseParallelStrategy.lie"""
        result = super(MeanParallelStrategy, self).lie(trial)
        if result:
            return result

        return Trial.Result(name='lie', type='lie', value=self.mean_result)


class StubParallelStrategy(BaseParallelStrategy):
    """Parallel strategy that returns static objective value for incompleted trials."""

    def __init__(self, stub_value=None):
        """Initialize the stub value"""
        self.stub_value = stub_value

    def observe(self, points, results):
        """See BaseParallelStrategy.observe"""
        pass

    def lie(self, trial):
        """See BaseParallelStrategy.lie"""
        result = super(StubParallelStrategy, self).lie(trial)
        if result:
            return result

        return Trial.Result(name='lie', type='lie', value=self.stub_value)


# pylint: disable=too-few-public-methods,abstract-method
class Strategy(BaseParallelStrategy, metaclass=Factory):
    """Class used to build a parallel strategy given name and params

    .. seealso:: `orion.core.utils.Factory` metaclass and `BaseParallelStrategy` interface.
    """

    pass



"""Collection of tests for :mod:`orion.plotting.orion.plotting.PlotAccessor`."""
import pytest

from orion.plotting.base import PlotAccessor
from orion.testing import create_experiment

config = dict(
    name='experiment-name',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'test-user',
              'orion_version': 'XYZ',
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir='',
    algorithms={'random': {'seed': 1}},
    producer={'strategy': 'NoParallelStrategy'},
)


trial_config = {
    'experiment': 0,
    'status': 'new',  # new, reserved, suspended, completed, broken
    'worker': None,
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [],
    'params': []
}


def check_regret_plot(plot):
    """Verifies that existence of the regret plot"""
    assert plot
    assert "regret" in plot.layout.title.text.lower()
    assert 2 == len(plot.data)


def test_init_require_experiment():
    """Tests that a `PlotAccessor` requires an instance of `ExperimentClient`"""
    with pytest.raises(ValueError) as exception:
        PlotAccessor(None)

    assert "Parameter 'experiment' is None" in str(exception.value)


def test_call_nonexistent_kind():
    """Tests that specifying a non existent kind will fail"""
    with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        with pytest.raises(ValueError) as exception:
            pa(kind='nonexistent')

        assert "Plot of kind 'nonexistent' is not one of" in str(exception.value)


def test_regret_is_default_plot():
    """Tests that the regret plot is the default plot"""
    with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        plot = pa()

        check_regret_plot(plot)


def test_regret_kind():
    """Tests that a regret plot can be created from specifying `kind` as a parameter."""
    with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        plot = pa(kind='regret')

        check_regret_plot(plot)


def test_call_to_regret():
    """Tests instance calls to `PlotAccessor.regret()`"""
    with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
        pa = PlotAccessor(experiment)
        plot = pa.regret()

        check_regret_plot(plot)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.db.test` -- Module to check if the DB works
=================================================================

.. module:: test_db
   :platform: Unix
   :synopsis: Runs multiple checks to see if the database was correctly setup.

"""
import argparse
import logging

from orion.core.cli.checks.creation import CreationStage
from orion.core.cli.checks.operations import OperationsStage
from orion.core.cli.checks.presence import PresenceStage
from orion.core.utils.exceptions import CheckError

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = 'Verifies that the database is correctly configured'


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    test_db_parser = parser.add_parser('test', help=SHORT_DESCRIPTION,
                                       description=SHORT_DESCRIPTION)

    test_db_parser.add_argument('-c', '--config', type=argparse.FileType('r'),
                                metavar='path-to-config', help="user provided "
                                "orion configuration file")

    test_db_parser.set_defaults(func=main)

    return test_db_parser


def main(args):
    """Run through all checks for database."""
    presence_stage = PresenceStage(args)
    creation_stage = CreationStage(presence_stage)
    operations_stage = OperationsStage(creation_stage)
    stages = [presence_stage, creation_stage, operations_stage]

    for stage in stages:
        for check in stage.checks():
            print(check.__doc__, end='.. ')
            try:
                status, msg = check()
                print(status)
                if status == "Skipping":
                    print(msg)
            except CheckError as ex:
                print("Failure")
                print(ex)

        stage.post_stage()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.hunt` -- Module running the optimization command
=====================================================================

.. module:: hunt
   :platform: Unix
   :synopsis: Gets an experiment and iterates over it until one of the exit conditions is met

"""

import logging

import orion.core
from orion.core.cli import base as cli
from orion.core.cli import evc as evc_cli
import orion.core.io.experiment_builder as experiment_builder
from orion.core.worker import workon

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = 'Conducts hyperparameter optimization'
DESCRIPTION = """
This command starts hyperparameter optimization process for the user-provided model using the
configured optimization algorithm and search space.
"""


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    hunt_parser = parser.add_parser('hunt', help=SHORT_DESCRIPTION, description=DESCRIPTION)

    orion_group = cli.get_basic_args_group(
        hunt_parser, group_name='Hunt arguments', group_help='')

    orion.core.config.experiment.add_arguments(
        orion_group,
        rename=dict(max_broken='--exp-max-broken', max_trials='--exp-max-trials'))

    orion_group.add_argument(
        '--max-trials', type=int, metavar='#',
        help="(DEPRECATED) This argument will be removed in v0.3. Use --exp-max-trials instead")

    orion_group.add_argument(
        '--init-only', default=False, action='store_true',
        help="Only create the experiment and register in database, but do not execute any trial.")

    worker_args_group = hunt_parser.add_argument_group(
        "Worker arguments (optional)",
        description="Arguments to automatically resolved branching events.")

    orion.core.config.worker.add_arguments(
        worker_args_group,
        rename=dict(max_broken='--worker-max-broken', max_trials='--worker-max-trials'))

    evc_cli.get_branching_args_group(hunt_parser)

    cli.get_user_args_group(hunt_parser)

    hunt_parser.set_defaults(func=main)
    hunt_parser.set_defaults(help_empty=True)  # Print help if command is empty

    return hunt_parser


def main(args):
    """Build experiment and execute hunt command"""
    args['root'] = None
    args['leafs'] = []
    # TODO: simplify when parameter parsing is refactored
    experiment = experiment_builder.build_from_args(args)

    if args['init_only']:
        return

    config = experiment_builder.get_cmd_config(args)
    worker_config = orion.core.config.worker.to_dict()
    if config.get('worker'):
        worker_config.update(config.get('worker'))

    workon(experiment, **worker_config)



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.singleton` -- Singleton helpers and boilerplate.
=======================================================================

.. module:: singleton
   :platform: Unix
   :synopsis: Helper functions and boilerplate for Singleton manipulation and creation.
"""
from abc import ABCMeta

from orion.core.utils import Factory


class SingletonAlreadyInstantiatedError(ValueError):
    """Exception to be raised when someone provides arguments to build
    an object from a already-instantiated `SingletonType` class.
    """

    def __init__(self, name):
        """Pass the same constant message to ValueError underneath."""
        super().__init__("A singleton instance of (type: {}) has already been instantiated."
                         .format(name))


class SingletonNotInstantiatedError(TypeError):
    """Exception to be raised when someone try to access an instance
    of a singleton that has not been instantiated yet
    """

    def __init__(self, name):
        """Pass the same constant message to TypeError underneath."""
        super().__init__('No singleton instance of (type: {}) was created'
                         .format(name))


class SingletonType(type):
    """Metaclass that implements the singleton pattern for a Python class."""

    def __init__(cls, name, bases, dictionary):
        """Create a class instance variable and initiate it to None object."""
        super(SingletonType, cls).__init__(name, bases, dictionary)
        cls.instance = None

    def __call__(cls, *args, **kwargs):
        """Create an object if does not already exist, otherwise return what there is."""
        if cls.instance is None:
            try:
                cls.instance = super(SingletonType, cls).__call__(*args, **kwargs)
            except TypeError as exception:
                raise SingletonNotInstantiatedError(cls.__name__) from exception

        elif args or kwargs:
            raise SingletonAlreadyInstantiatedError(cls.__name__)

        return cls.instance


class AbstractSingletonType(SingletonType, ABCMeta):
    """This will create singleton base classes, that need to be subclassed and implemented."""

    pass


class SingletonFactory(AbstractSingletonType, Factory):
    """Wrapping `Factory` with `SingletonType`. Keep compatibility with `AbstractSingletonType`."""

    pass


def update_singletons(values=None):
    """Replace singletons by given values and return previous singleton objects"""
    if values is None:
        values = {}

    # Avoiding circular import problems when importing this module.
    from orion.core.io.database import Database
    from orion.core.io.database.ephemeraldb import EphemeralDB
    from orion.core.io.database.mongodb import MongoDB
    from orion.core.io.database.pickleddb import PickledDB
    from orion.storage.base import Storage
    from orion.storage.legacy import Legacy
    from orion.storage.track import Track

    singletons = (Storage, Legacy, Database, MongoDB, PickledDB, EphemeralDB, Track)

    updated_singletons = {}
    for singleton in singletons:
        updated_singletons[singleton] = singleton.instance
        singleton.instance = values.get(singleton, None)

    return updated_singletons



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the default commands for demo purposes."""
import os

import pytest

import orion.core.cli
from orion.storage.base import get_storage


def get_user_corneau():
    """Return user corneau (to mock getpass.getuser)"""
    return "corneau"


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_insert_invalid_experiment(database, monkeypatch, capsys):
    """Test the insertion of an invalid experiment"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    returncode = orion.core.cli.main(
        ["insert", "-n", "dumb_experiment",
         "-c", "./orion_config_random.yaml", "./black_box.py", "-x=1"])

    assert returncode == 1

    captured = capsys.readouterr().err

    assert ("Error: No experiment with given name 'dumb_experiment'")


@pytest.mark.usefixtures("only_experiments_db")
@pytest.mark.usefixtures("null_db_instances")
def test_insert_single_trial(database, monkeypatch, script_path):
    """Try to insert a single trial"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    orion.core.cli.main(["insert", "-n", "test_insert_normal",
                         "-c", "./orion_config_random.yaml", script_path, "-x=1"])

    exp = list(database.experiments.find({"name": "test_insert_normal"}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp

    trials = list(database.trials.find({"experiment": exp['_id']}))

    assert len(trials) == 1

    trial = trials[0]

    assert trial['status'] == 'new'
    assert trial['params'][0]['value'] == 1


@pytest.mark.usefixtures("only_experiments_db")
@pytest.mark.usefixtures("null_db_instances")
def test_insert_single_trial_default_value(database, monkeypatch):
    """Try to insert a single trial using a default value"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    orion.core.cli.main(["insert", "-n", "test_insert_normal",
                         "-c", "./orion_config_random.yaml", "./black_box.py"])

    exp = list(database.experiments.find({"name": "test_insert_normal"}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp

    trials = list(database.trials.find({"experiment": exp['_id']}))

    assert len(trials) == 1

    trial = trials[0]

    assert trial['status'] == 'new'
    assert trial['params'][0]['value'] == 1


@pytest.mark.usefixtures("only_experiments_db")
@pytest.mark.usefixtures("null_db_instances")
def test_insert_with_no_default_value(database, monkeypatch):
    """Try to insert a single trial by omitting a namespace with no default value"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    with pytest.raises(ValueError) as exc_info:
        orion.core.cli.main(["insert", "-n", "test_insert_missing_default_value",
                             "-c", "./orion_config_random.yaml", "./black_box.py"])

    assert "Dimension /x is unspecified and has no default value" in str(exc_info.value)


@pytest.mark.usefixtures("only_experiments_db")
@pytest.mark.usefixtures("null_db_instances")
def test_insert_with_incorrect_namespace(database, monkeypatch):
    """Try to insert a single trial with a namespace not inside the experiment space"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    with pytest.raises(ValueError) as exc_info:
        orion.core.cli.main(["insert", "-n", "test_insert_normal",
                             "-c", "./orion_config_random.yaml",
                             "./black_box.py", "-p=4"])

    assert "Found namespace outside of experiment space : /p" in str(exc_info.value)


@pytest.mark.usefixtures("only_experiments_db")
@pytest.mark.usefixtures("null_db_instances")
def test_insert_with_outside_bound_value(database, monkeypatch):
    """Try to insert a single trial with value outside the distribution's interval"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)

    with pytest.raises(ValueError) as exc_info:
        orion.core.cli.main(["insert", "-n", "test_insert_two_hyperparameters",
                             "-c", "./orion_config_random.yaml",
                             "./black_box.py", "-x=4", "-y=100"])
    assert "Value 100 is outside of" in str(exc_info.value)


@pytest.mark.usefixtures("only_experiments_db")
@pytest.mark.usefixtures("null_db_instances")
def test_insert_two_hyperparameters(database, monkeypatch):
    """Try to insert a single trial with two hyperparameters"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setattr("getpass.getuser", get_user_corneau)
    orion.core.cli.main(["insert", "-n", "test_insert_two_hyperparameters",
                         "-c", "./orion_config_random.yaml", "./black_box.py", "-x=1", "-y=2"])

    exp = list(database.experiments.find({"name": "test_insert_two_hyperparameters"}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp

    trials = list(database.trials.find({"experiment": exp['_id']}))

    assert len(trials) == 1

    trial = trials[0]

    assert trial['status'] == 'new'
    assert trial['params'][0]['value'] == 1
    assert trial['params'][1]['value'] == 2


@pytest.mark.usefixtures("clean_db")
def test_insert_with_version(create_db_instance, monkeypatch, script_path):
    """Try to insert a single trial inside a specific version"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["hunt", "--init-only", "-n", "experiment",
                         "-c", "./orion_config_random.yaml", script_path, "-x~normal(0,1)"])
    orion.core.cli.main(["hunt", "--init-only", "-n", "experiment",
                         "-c", "./orion_config_random.yaml", script_path, "-x~normal(0,1)",
                         "-y~+normal(0,1)"])

    exp = list(get_storage().fetch_experiments({"name": "experiment", "version": 1}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp

    trials = list(get_storage().fetch_trials(uid=exp['_id']))
    assert len(trials) == 0

    orion.core.cli.main(["insert", "-n", "experiment", "--version", "1",
                         "-c", "./orion_config_random.yaml", script_path, "-x=1"])

    trials = list(get_storage().fetch_trials(uid=exp['_id']))

    assert len(trials) == 1


def test_no_args(capsys):
    """Test that help is printed when no args are given."""
    with pytest.raises(SystemExit):
        orion.core.cli.main(['insert'])

    captured = capsys.readouterr().out

    assert 'usage:' in captured
    assert 'Traceback' not in captured


def test_no_name(capsys):
    """Try to run the command without providing an experiment name"""
    returncode = orion.core.cli.main(["insert", "--version", "1"])
    assert returncode == 1

    captured = capsys.readouterr().err

    assert captured == 'Error: No name provided for the experiment.\n'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.orion_cmdliner_parser`."""
import os

import pytest

from orion.core.io.orion_cmdline_parser import OrionCmdlineParser
from orion.core.worker.trial import Trial


@pytest.fixture
def script_path():
    """Return existing script path for the command lines"""
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def parser():
    """Return an instance of `OrionCmdlineParser`."""
    return OrionCmdlineParser(allow_non_existing_files=True)


@pytest.fixture
def parser_diff_prefix():
    """Return an instance of `OrionCmdlineParser` with a different config prefix."""
    return OrionCmdlineParser(config_prefix='config2', allow_non_existing_files=True)


@pytest.fixture
def commandline():
    """Return a simple commandline list."""
    return ["--seed=555",
            "--lr~uniform(-3, 1)",
            "--non-prior=choices({{'sgd': 0.2, 'adam': 0.8}})",
            "--prior~choices({'sgd': 0.2, 'adam': 0.8})"]


@pytest.fixture
def commandline_fluff(commandline):
    """Add extra useless info to commandline."""
    cmd_args = commandline
    cmd_args.extend(["--some-path=~/some_path",
                     "--home-path=~",
                     "~/../folder/.hidden/folder"])
    return cmd_args


@pytest.fixture
def cmd_with_properties(commandline):
    """Add extra arguments that use `trial` and `exp` properties"""
    cmd_args = commandline
    cmd_args.extend(["--trial-name", "{trial.hash_name}", "--exp-name", "{exp.name}"])

    return cmd_args


def test_parse_from_yaml_config(parser, yaml_config):
    """Parse from a yaml config only."""
    parser.parse(yaml_config)
    config = parser.priors

    assert len(config.keys()) == 6
    assert '/layers/1/width' in config
    assert '/layers/1/type' in config
    assert '/layers/2/type' in config
    assert '/training/lr0' in config
    assert '/training/mbs' in config
    assert '/something-same' in config


def test_parse_from_json_config(parser, json_config):
    """Parse from a json config only."""
    parser.parse(json_config)
    config = parser.priors

    assert len(config.keys()) == 6
    assert '/layers/1/width' in config
    assert '/layers/1/type' in config
    assert '/layers/2/type' in config
    assert '/training/lr0' in config
    assert '/training/mbs' in config
    assert '/something-same' in config


def test_parse_from_unknown_config(parser, some_sample_config):
    """Parse from a unknown config type only."""
    parser.parse(some_sample_config)
    config = parser.priors

    assert len(config.keys()) == 6
    assert '/layers/1/width' in config
    assert '/layers/1/type' in config
    assert '/layers/2/type' in config
    assert '/training/lr0' in config
    assert '/training/mbs' in config
    assert '/something-same' in config


def test_parse_equivalency(yaml_config, json_config):
    """Templates found from json and yaml are the same."""
    parser_yaml = OrionCmdlineParser(allow_non_existing_files=True)
    parser_yaml.parse(yaml_config)
    dict_from_yaml = parser_yaml.config_file_data

    parser_json = OrionCmdlineParser(allow_non_existing_files=True)
    parser_json.parse(json_config)
    dict_from_json = parser_json.config_file_data
    assert dict_from_json == dict_from_yaml


def test_parse_from_args_only(parser, commandline_fluff):
    """Parse a commandline."""
    cmd_args = commandline_fluff

    parser.parse(cmd_args)

    assert not parser.config_file_data
    assert len(parser.cmd_priors) == 2
    assert '/lr' in parser.cmd_priors
    assert '/prior' in parser.cmd_priors
    assert parser.parser.template == \
        ['--seed', '{seed}', '--lr', '{lr}', '--non-prior', '{non-prior}', '--prior', '{prior}',
         '--some-path', '{some-path}', '--home-path', '{home-path[0]}', '{home-path[1]}']


def test_parse_from_args_and_config_yaml(parser, commandline, yaml_config):
    """Parse both from commandline and config file."""
    cmd_args = yaml_config
    cmd_args.extend(commandline)

    parser.parse(cmd_args)
    config = parser.priors

    assert len(config) == 8
    assert '/lr' in config
    assert '/prior' in config
    assert '/layers/1/width' in config
    assert '/layers/1/type' in config
    assert '/layers/2/type' in config
    assert '/training/lr0' in config
    assert '/training/mbs' in config
    assert '/something-same' in config

    template = parser.parser.template
    assert template == ['--config', '{config}', '--seed', '{seed}', '--lr', '{lr}',
                        '--non-prior', '{non-prior}', '--prior', '{prior}']


def test_parse_finds_conflict(parser, commandline, yaml_config):
    """Parse find conflicting declaration in commandline and config file."""
    cmd_args = yaml_config
    cmd_args.extend(commandline)
    cmd_args.extend(["--something-same~choices({'sgd': 0.2, 'adam': 0.8})"])

    with pytest.raises(ValueError) as exc:
        parser.parse(cmd_args)

    assert 'Conflict' in str(exc.value)


def test_format_commandline_only(parser, commandline):
    """Format the commandline using only args."""
    parser.parse(commandline)

    trial = Trial(params=[
        {'name': '/lr', 'type': 'real', 'value': -2.4},
        {'name': '/prior', 'type': 'categorical', 'value': 'sgd'}])

    cmd_inst = parser.format(trial=trial)
    assert cmd_inst == ["--seed", "555",
                        "--lr", "-2.4",
                        "--non-prior", "choices({'sgd': 0.2, 'adam': 0.8})",
                        "--prior", "sgd"]


def test_format_commandline_and_config(parser, commandline, json_config, tmpdir, json_converter):
    """Format the commandline and a configuration file."""
    cmd_args = json_config
    cmd_args.extend(commandline)

    parser.parse(cmd_args)

    trial = Trial(params=[
        {'name': '/lr', 'type': 'real', 'value': -2.4},
        {'name': '/prior', 'type': 'categorical', 'value': 'sgd'},
        {'name': '/layers/1/width', 'type': 'integer', 'value': 100},
        {'name': '/layers/1/type', 'type': 'categorical', 'value': 'relu'},
        {'name': '/layers/2/type', 'type': 'categorical', 'value': 'sigmoid'},
        {'name': '/training/lr0', 'type': 'real', 'value': 0.032},
        {'name': '/training/mbs', 'type': 'integer', 'value': 64},
        {'name': '/something-same', 'type': 'categorical', 'value': '3'}])

    output_file = str(tmpdir.join("output.json"))

    cmd_inst = parser.format(output_file, trial)

    assert cmd_inst == ['--config', output_file, "--seed", "555", "--lr", "-2.4",
                        "--non-prior", "choices({'sgd': 0.2, 'adam': 0.8})", "--prior", "sgd"]

    output_data = json_converter.parse(output_file)
    assert output_data == {'yo': 5, 'training': {'lr0': 0.032, 'mbs': 64},
                           'layers': [{'width': 64, 'type': 'relu'},
                                      {'width': 100, 'type': 'relu'},
                                      {'width': 16, 'type': 'sigmoid'}],
                           'something-same': '3'}


def test_format_without_config_path(parser, commandline, json_config, tmpdir, json_converter):
    """Verify that parser.format() raises ValueError when config path not passed."""
    cmd_args = json_config
    cmd_args.extend(commandline)

    parser.parse(cmd_args)

    trial = Trial(params=[
        {'name': '/lr', 'type': 'real', 'value': -2.4},
        {'name': '/prior', 'type': 'categorical', 'value': 'sgd'},
        {'name': '/layers/1/width', 'type': 'integer', 'value': 100},
        {'name': '/layers/1/type', 'type': 'categorical', 'value': 'relu'},
        {'name': '/layers/2/type', 'type': 'categorical', 'value': 'sigmoid'},
        {'name': '/training/lr0', 'type': 'real', 'value': 0.032},
        {'name': '/training/mbs', 'type': 'integer', 'value': 64},
        {'name': '/something-same', 'type': 'categorical', 'value': '3'}])

    with pytest.raises(ValueError) as exc_info:
        parser.format(trial=trial)

    assert "Cannot format without a `config_path` argument." in str(exc_info.value)


def test_format_with_properties(parser, cmd_with_properties, hacked_exp):
    """Test if format correctly puts the value of `trial` and `exp` when used as properties"""
    parser.parse(cmd_with_properties)

    trial = Trial(experiment='trial_test', params=[
        {'name': '/lr', 'type': 'real', 'value': -2.4},
        {'name': '/prior', 'type': 'categorical', 'value': 'sgd'}])

    cmd_line = parser.format(None, trial=trial, experiment=hacked_exp)

    assert trial.hash_name in cmd_line
    assert 'supernaedo2-dendi' in cmd_line


def test_configurable_config_arg(parser_diff_prefix, yaml_sample_path):
    """Parse from a yaml config only."""
    parser_diff_prefix.parse(["--config2", yaml_sample_path])
    config = parser_diff_prefix.priors

    assert len(config.keys()) == 6
    assert '/layers/1/width' in config
    assert '/layers/1/type' in config
    assert '/layers/2/type' in config
    assert '/training/lr0' in config
    assert '/training/mbs' in config
    assert '/something-same' in config


def test_infer_user_script(script_path):
    """Test that user script is infered correctly"""
    parser = OrionCmdlineParser()
    parser.parse(f'{script_path} and some args'.split(' '))
    assert parser.user_script == script_path


def test_infer_user_script_python(script_path):
    """Test that user script is infered correctly when using python"""
    parser = OrionCmdlineParser()
    parser.parse(f'python {script_path} and some args'.split(' '))
    assert parser.user_script == script_path


def test_infer_user_script_when_missing():
    """Test that user script is infered correctly even if missing"""
    parser = OrionCmdlineParser()

    with pytest.raises(FileNotFoundError) as exc:
        parser.parse('python script.py and some args'.split(' '))
    assert exc.match('The path specified for the script does not exist')

    parser = OrionCmdlineParser(allow_non_existing_files=True)
    parser.parse('python script.py and some args'.split(' '))
    assert parser.user_script == 'script.py'


def test_configurable_config_arg_do_not_exist(script_path):
    """Test that parser can handle command if config file does not exist"""
    parser = OrionCmdlineParser()
    command = f'python {script_path} --config idontexist.yaml'.split(' ')
    with pytest.raises(OSError) as exc:
        parser.parse(command)
    assert exc.match("The path specified for the script config does not exist")

    parser = OrionCmdlineParser(allow_non_existing_files=True)
    parser.parse(command)


def test_get_state_dict_before_parse(parser, commandline):
    """Test getting state dict."""
    assert parser.get_state_dict() == {
        'parser': {
            'keys': [],
            'arguments': [],
            'template': []},
        'cmd_priors': list(map(list, parser.cmd_priors.items())),
        'file_priors': list(map(list, parser.file_priors.items())),
        'config_file_data': parser.config_file_data,
        'config_prefix': parser.config_prefix,
        'file_config_path': parser.file_config_path,
        'converter': None}


def test_get_state_dict_after_parse_no_config_file(parser, commandline):
    """Test getting state dict."""
    parser.parse(commandline)

    assert parser.get_state_dict() == {
        'parser': parser.parser.get_state_dict(),
        'cmd_priors': list(map(list, parser.cmd_priors.items())),
        'file_priors': list(map(list, parser.file_priors.items())),
        'config_file_data': parser.config_file_data,
        'config_prefix': parser.config_prefix,
        'file_config_path': parser.file_config_path,
        'converter': None}


def test_get_state_dict_after_parse_with_config_file(parser, yaml_config, commandline):
    """Test getting state dict."""
    cmd_args = yaml_config
    cmd_args.extend(commandline)

    parser.parse(cmd_args)

    assert parser.get_state_dict() == {
        'parser': parser.parser.get_state_dict(),
        'cmd_priors': list(map(list, parser.cmd_priors.items())),
        'file_priors': list(map(list, parser.file_priors.items())),
        'config_file_data': parser.config_file_data,
        'config_prefix': parser.config_prefix,
        'file_config_path': parser.file_config_path,
        'converter': parser.converter.get_state_dict()}


def test_set_state_dict(parser, commandline, json_config, tmpdir, json_converter):
    """Test that set_state_dict sets state properly to generate new config."""
    cmd_args = json_config
    cmd_args.extend(commandline)

    parser.parse(cmd_args)

    state = parser.get_state_dict()
    parser = None

    blank_parser = OrionCmdlineParser(allow_non_existing_files=True)

    blank_parser.set_state_dict(state)

    trial = Trial(params=[
        {'name': '/lr', 'type': 'real', 'value': -2.4},
        {'name': '/prior', 'type': 'categorical', 'value': 'sgd'},
        {'name': '/layers/1/width', 'type': 'integer', 'value': 100},
        {'name': '/layers/1/type', 'type': 'categorical', 'value': 'relu'},
        {'name': '/layers/2/type', 'type': 'categorical', 'value': 'sigmoid'},
        {'name': '/training/lr0', 'type': 'real', 'value': 0.032},
        {'name': '/training/mbs', 'type': 'integer', 'value': 64},
        {'name': '/something-same', 'type': 'categorical', 'value': '3'}])

    output_file = str(tmpdir.join("output.json"))

    cmd_inst = blank_parser.format(output_file, trial)

    assert cmd_inst == ['--config', output_file, "--seed", "555", "--lr", "-2.4",
                        "--non-prior", "choices({'sgd': 0.2, 'adam': 0.8})", "--prior", "sgd"]

    output_data = json_converter.parse(output_file)
    assert output_data == {'yo': 5, 'training': {'lr0': 0.032, 'mbs': 64},
                           'layers': [{'width': 64, 'type': 'relu'},
                                      {'width': 100, 'type': 'relu'},
                                      {'width': 16, 'type': 'sigmoid'}],
                           'something-same': '3'}



"""Perform tests for the REST endpoint `/trials`"""
import copy
import datetime
import random

from orion.core.io import experiment_builder
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage

base_experiment = dict(
    name='experiment-name',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'test-user',
              'orion_version': 'x.y.z',
              'datetime': datetime.datetime(1, 1, 1),
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    refers={},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir='',
    algorithms={'random': {'seed': 1}},
)

base_trial = {
    'experiment': None,
    'status': 'new',
    'worker': None,
    'submit_time': datetime.datetime(1, 1, 1),
    'start_time': datetime.datetime(1, 1, 1, second=10),
    'end_time': datetime.datetime(1, 1, 2),
    'heartbeat': None,
    'results': [
        {'name': 'loss', 'type': 'objective', 'value': 0.05},
        {'name': 'a', 'type': 'statistic', 'value': 10},
        {'name': 'b', 'type': 'statistic', 'value': 5},
    ],
    'params': [
        {'name': 'x',
         'type': 'real',
         'value': 10.0},
    ],
}


def add_experiment(**kwargs):
    """Adds experiment to the dummy orion instance"""
    base_experiment.update(copy.deepcopy(kwargs))
    experiment_builder.build(branching=dict(branch_from=base_experiment['name']), **base_experiment)


def add_trial(experiment: int, status: str = None, **kwargs):
    """
    Add trials to the dummy orion instance

    Parameters
    ----------
    experiment
        The ID of the experiment (stored in Experiment._id)
    status
        The status of the trial to add.
    **kwargs
        Any other attribute to modify from the base trial config.
    """
    if not status:
        status = random.choice(Trial.allowed_stati)

    kwargs['experiment'] = experiment
    kwargs['status'] = status

    base_trial.update(copy.deepcopy(kwargs))
    get_storage().register_trial(Trial(**base_trial))


def test_root_endpoint_not_supported(client):
    """Tests that the server return a 404 when accessing trials/ with no experiment"""
    response = client.simulate_get('/trials')

    assert response.status == "404 Not Found"
    assert not response.json


class TestTrialCollection:
    """Tests trials/:experiment_name"""

    def test_trials_for_unknown_experiment(self, client):
        """Tests that an unknown experiment returns a not found error"""
        response = client.simulate_get('/trials/unknown-experiment')

        assert response.status == "404 Not Found"
        assert response.json == {'title': 'Experiment not found',
                                 'description': 'Experiment "unknown-experiment" does not exist'}

    def test_unknown_parameter(self, client):
        """
        Tests that if an unknown parameter is specified in
        the query string, an error is returned even if the experiment doesn't exist.
        """
        expected_error_message = 'Parameter "unknown" is not supported. ' \
                                 'Expected one of [\'ancestors\', \'status\', \'version\'].'

        response = client.simulate_get('/trials/a?unknown=true')

        assert response.status == "400 Bad Request"
        assert response.json == {'title': 'Invalid parameter',
                                 'description': expected_error_message}

        add_experiment(name='a', version=1, _id=1)

        response = client.simulate_get('/trials/a?unknown=true')

        assert response.status == "400 Bad Request"
        assert response.json == {'title': 'Invalid parameter',
                                 'description': expected_error_message}

    def test_trials_for_latest_version(self, client):
        """Tests that it returns the trials of the latest version of the experiment"""
        add_experiment(name='a', version=1, _id=1)
        add_experiment(name='a', version=2, _id=2)

        add_trial(experiment=1, id_override="00")
        add_trial(experiment=2, id_override="01")
        add_trial(experiment=1, id_override="02")
        add_trial(experiment=2, id_override="03")

        response = client.simulate_get('/trials/a')

        assert response.status == "200 OK"
        assert response.json == [{'id': '01'}, {'id': '03'}]

    def test_trials_for_specific_version(self, client):
        """Tests specific version of experiment"""
        add_experiment(name='a', version=1, _id=1)
        add_experiment(name='a', version=2, _id=2)
        add_experiment(name='a', version=3, _id=3)

        add_trial(experiment=1, id_override="00")
        add_trial(experiment=2, id_override="01")
        add_trial(experiment=3, id_override="02")

        # Happy case
        response = client.simulate_get('/trials/a?version=2')

        assert response.status == "200 OK"
        assert response.json == [{'id': '01'}]

        # Version doesn't exist
        response = client.simulate_get('/trials/a?version=4')

        assert response.status == "404 Not Found"
        assert response.json == {
            'title': 'Experiment not found',
            'description': 'Experiment "a" has no version "4"'
        }

    def test_trials_for_all_versions(self, client):
        """Tests that trials from all ancestors are shown"""
        add_experiment(name='a', version=1, _id=1)
        add_experiment(name='a', version=2, _id=2)
        add_experiment(name='a', version=3, _id=3)

        add_trial(experiment=1, id_override="00")
        add_trial(experiment=2, id_override="01")
        add_trial(experiment=3, id_override="02")

        # Happy case default
        response = client.simulate_get('/trials/a?ancestors=true')

        assert response.status == "200 OK"
        assert response.json == [{'id': '00'}, {'id': '01'}, {'id': '02'}]

        # Happy case explicitly false (call latest version test)
        response = client.simulate_get('/trials/a?ancestors=false')

        assert response.status == "200 OK"
        assert response.json == [{'id': '02'}]

        # Not a boolean parameter
        response = client.simulate_get('/trials/a?ancestors=42')

        assert response.status == "400 Bad Request"
        assert response.json == {
            'title': 'Invalid parameter',
            'description': 'The "ancestors" parameter is invalid. '
                           'The value of the parameter must be "true" or "false".'
        }

    def test_trials_by_status(self, client):
        """Tests that trials are returned"""
        add_experiment(name='a', version=1, _id=1)

        # There exist no trial of the given status in an empty experiment
        response = client.simulate_get('/trials/a?status=completed')

        assert response.status == "200 OK"
        assert response.json == []

        # There exist no trial of the given status while other status are present
        add_trial(experiment=1, id_override="00", status="broken")

        response = client.simulate_get('/trials/a?status=completed')
        assert response.status == "200 OK"
        assert response.json == []

        # There exist at least one trial of the given status
        add_trial(experiment=1, id_override="01", status="completed")

        response = client.simulate_get('/trials/a?status=completed')

        assert response.status == "200 OK"
        assert response.json == [{'id': '01'}]

        # Status doesn't exist
        response = client.simulate_get('/trials/a?status=invalid')

        assert response.status == "400 Bad Request"
        error_message = "Invalid status value. " \
                        "Expected one of " \
                        "['new', 'reserved', 'suspended', 'completed', 'interrupted', 'broken']"
        assert response.json == {'title': 'Invalid parameter',
                                 'description': "The \"status\" parameter is invalid. "
                                                "The value of the parameter must be one of "
                                                "['new', 'reserved', 'suspended', 'completed', "
                                                "'interrupted', 'broken']"}

    def test_trials_by_from_specific_version_by_status_with_ancestors(self, client):
        """Tests that mixing parameters work as intended"""
        add_experiment(name='a', version=1, _id=1)
        add_experiment(name='b', version=1, _id=2)
        add_experiment(name='a', version=2, _id=3)
        add_experiment(name='a', version=3, _id=4)

        add_trial(experiment=1, id_override="00", status='completed')
        add_trial(experiment=3, id_override="01", status='broken')
        add_trial(experiment=3, id_override="02", status='completed')
        add_trial(experiment=2, id_override="03", status='completed')

        response = client.simulate_get('/trials/a?ancestors=true&version=2&status=completed')

        assert response.status == "200 OK"
        assert response.json == [{'id': '00'}, {'id': '02'}]


class TestTrialItem:
    """Tests trials/:experiment_name/:trial_id"""

    def test_unknown_experiment(self, client):
        """Tests that an unknown experiment returns a not found error"""
        response = client.simulate_get('/trials/unknown-experiment/a-trial')

        assert response.status == "404 Not Found"
        assert response.json == {'title': 'Experiment not found',
                                 'description': 'Experiment "unknown-experiment" does not exist'}

    def test_unknown_trial(self, client):
        """Tests that an unknown experiment returns a not found error"""
        add_experiment(name='a', version=1, _id=1)

        response = client.simulate_get('/trials/a/unknown-trial')

        assert response.status == "404 Not Found"
        assert response.json == {'title': 'Trial not found',
                                 'description': 'Trial "unknown-trial" does not exist'}

    def test_get_trial(self, client):
        """Tests that an existing trial is returned according to the API specification"""
        add_experiment(name='a', version=1, _id=1)
        add_trial(experiment=1, id_override="00", status='completed')
        add_trial(experiment=1, id_override="01", status='completed')

        response = client.simulate_get('/trials/a/01')

        assert response.status == "200 OK"
        assert response.json == {
            'id': '01',
            'submitTime': '0001-01-01 00:00:00',
            'startTime': '0001-01-01 00:00:10',
            'endTime': '0001-01-02 00:00:00',
            'parameters': {'x': 10.0},
            'objective': 0.05,
            'statistics': {'a': 10,
                           'b': 5},
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.checks.presence` -- Presence stage for database checks
===========================================================================

.. module:: presence
    :platform: Unix
    :synopsis: Checks for the presence of a configuration.

"""

import orion.core
import orion.core.io.experiment_builder as experiment_builder
import orion.core.utils.backward as backward


class PresenceStage:
    """The presence stage of the checks."""

    def __init__(self, cmdargs):
        """Create an instance of the stage."""
        self.cmdargs = cmdargs
        self.db_config = {}

    def checks(self):
        """Return the registered checks."""
        yield self.check_default_config
        yield self.check_configuration_file

    def check_default_config(self):
        """Check for a configuration inside the default paths."""
        config = orion.core.config.to_dict()

        backward.update_db_config(config)

        if 'database' not in config.get('storage', {}):
            return "Skipping", "No default configuration found for database."

        self.db_config = config['storage']['database']
        print('\n   ', self.db_config)

        return "Success", ""

    def check_configuration_file(self):
        """Check if configuration file has valid database configuration."""
        config = experiment_builder.get_cmd_config(self.cmdargs)

        if not len(config):
            return "Skipping", "Missing configuration file."

        if 'database' not in config.get('storage', {}):
            return "Skipping", "No database found in configuration file."

        config = config['storage']['database']
        names = ['type', 'name', 'host', 'port']

        if not any(name in config for name in names):
            return "Skipping", "No configuration value found inside `database`."

        self.db_config.update(config)

        print('\n   ', config)

        return "Success", ""

    def post_stage(self):
        """Print the current config."""
        print("Using configuration: {}".format(self.db_config))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.interactive_commands.base_commands` -- CLI for conflict solving
===================================================================================
.. module:: base_commands
    :platform: Unix
    :synopsis: Launch the interactive prompt and call the related commands

.... ....  TODO
"""

import argparse
import cmd
import functools
import io
import os
import readline
import shlex
import traceback

from orion.algo.space import Dimension
from orion.core.evc import adapters, conflicts
from orion.core.utils.diff import green, red
from orion.storage.base import get_storage


readline.set_completer_delims(' ')


def wrap_autocomplete(f):
    """Wrap autocomplete to catch errors and print stacktrace"""
    @functools.wraps(f)
    def call(self, *args):

        # Don't know why, it gets reset to default ones between 2 auto-completions
        readline.set_completer_delims(" ")

        try:
            completions = f(self, *args)
        except KeyboardInterrupt:
            raise
        except BaseException:
            print(traceback.format_exc())
            completions = []

        return completions

    return call


def parse_command(f):
    """Wrap command methods to automatically parse using parsers and catch errors to print
    stacktrace without leaving the prompt
    """
    @functools.wraps(f)
    def call(self, arg):

        try:
            options = self.parser.parse_args([f.__name__[3:]] + shlex.split(arg))
        except SystemExit:
            return None

        try:
            rval = f(self, options)
        except KeyboardInterrupt:
            raise
        except BaseException:
            print(traceback.format_exc())
            rval = None

        return rval

    return call


class BranchingPrompt(cmd.Cmd):
    """
    Interactive command prompt to solve the configurations conflicts
    between the parent configuration and the new one.
    """

    intro = ('\n\n'
             'Welcome to Orion\'s experiment branching interactive conflicts resolver\n'
             '-----------------------------------------------------------------------\n\n'
             'If you are unfamiliar with this process, you can type '
             '`help` to print the help message. You can also type `abort` or `(q)uit` at any '
             'moment to quit without saving.\n'
             '\n'
             '%s')
    prompt = '(orion) '

    def __init__(self, branch_builder):
        """Retrieve the instance of ExperimentBranchBuilder containing the conflicts"""
        cmd.Cmd.__init__(self)
        self.branch_builder = branch_builder
        self.abort = False
        self._build_parsers()

    def _build_parsers(self):
        self.parser = argparse.ArgumentParser(prog="(orion)", add_help=False)
        subparsers = self.parser.add_subparsers(title="commands")
        for command in self.get_commands():
            subparser_builder = getattr(self, '_add_{}_parser'.format(command), None)
            if subparser_builder:
                subparser_builder(subparsers)
            else:
                self._add_default_subparser(command, subparsers)

    def _build_argument_parser(self, subparsers, name, **kwargs):
        return subparsers.add_parser(name, **kwargs)

    def _add_default_subparser(self, name, subparsers):
        method = getattr(self, "do_{}".format(name))
        subparser = self._build_argument_parser(subparsers, name,
                                                help=getattr(method, "__doc__", None),
                                                description=getattr(method, "__doc__", None))
        return subparser

    def _add_name_parser(self, subparsers):
        subparser = self._build_argument_parser(subparsers, 'name', help=self.do_name.__doc__)
        subparser.add_argument('experiment_name', metavar='experiment-name')
        return subparser

    def _add_code_parser(self, subparsers):
        subparser = self._build_argument_parser(subparsers, 'code', help=self.do_code.__doc__)
        subparser.add_argument('change_type', choices=adapters.CodeChange.types)
        return subparser

    def _add_commandline_parser(self, subparsers):
        subparser = self._build_argument_parser(subparsers, 'commandline',
                                                help=self.do_commandline.__doc__)
        subparser.add_argument('change_type', choices=adapters.CommandLineChange.types)
        return subparser

    def _add_config_parser(self, subparsers):
        subparser = self._build_argument_parser(subparsers, 'config', help=self.do_config.__doc__)
        subparser.add_argument('change_type', choices=adapters.ScriptConfigChange.types)
        return subparser

    def _add_add_parser(self, subparsers):
        subparser = self._build_argument_parser(subparsers, 'add', help=self.do_add.__doc__)
        subparser.add_argument('dimension_name', metavar='dimension-name')
        subparser.add_argument('--default-value', default=Dimension.NO_DEFAULT_VALUE)
        return subparser

    def _add_remove_parser(self, subparsers):
        subparser = self._build_argument_parser(subparsers, 'remove', help=self.do_remove.__doc__)
        subparser.add_argument('dimension_name', metavar='dimension-name')
        subparser.add_argument('--default-value', default=Dimension.NO_DEFAULT_VALUE)
        return subparser

    def _add_rename_parser(self, subparsers):
        subparser = self._build_argument_parser(subparsers, 'rename', help=self.do_rename.__doc__)
        subparser.add_argument('old_name', metavar='old-name')
        subparser.add_argument('new_name', metavar='new-name')
        return subparser

    def _add_reset_parser(self, subparsers):
        subparser = self._build_argument_parser(subparsers, 'reset', help=self.do_reset.__doc__)
        subparser.add_argument('resolutions', nargs='+', metavar='resolutions')
        return subparser

    def solve_conflicts(self):
        """Start the prompt for the interactive conflicts solving"""
        self.cmdloop()

    def get_commands(self):
        """Get command names of the prompt"""
        # Get rid of 'do_'
        commands = [attr[3:] for attr in dir(self)
                    if callable(getattr(self, attr)) and attr.startswith('do_')]

        return commands

    def cmdloop(self):
        """Start cmdloop of the prompt"""
        intro = self.intro % self.get_status()
        super(BranchingPrompt, self).cmdloop(intro)

    def get_status(self, options=None):
        """Return a string representing the status"""
        unsolved_conflicts = self.branch_builder.conflicts.get_remaining()
        resolved_conflicts = self.branch_builder.conflicts.get_resolved()

        output = io.StringIO()
        if resolved_conflicts:
            resolution_strings = set(str(conflict.resolution) for conflict in resolved_conflicts)
            print('Resolutions:', file=output)
            print(file=output)
            for resolution_string in resolution_strings:
                print("    ", green(str(resolution_string)), file=output)
            print(file=output)

            if unsolved_conflicts:
                print(file=output)

        if unsolved_conflicts:

            print('Remaining conflicts:', file=output)
            print(file=output)
            for conflict in unsolved_conflicts:
                print("    ", red(str(conflict)), file=output)
            print(file=output)
        else:
            print(file=output)
            print("Hooray, there is no more conflicts!", file=output)
            print("You can enter \'commit\' to leave this prompt and register the new branch",
                  file=output)
            print(file=output)

        return output.getvalue()

    # Commands
    @parse_command
    def do_help(self, options):
        """Print help message for all commands"""
        output = io.StringIO(initial_value='', newline='\n')
        self.parser.print_help(file=output)
        print(output.getvalue())

    @parse_command
    def do_h(self, options):
        """Alias for help"""
        self.do_help("")

    def do_shell(self, line):
        """Run a shell command. Ex: (orion) ! pwd"""
        print("running shell command:", line)
        print(os.popen(line).read())

    @parse_command
    def do_status(self, options):
        """Display the current status of the conflicting configuration"""
        print()
        print(self.get_status(options))

    @parse_command
    def do_diff(self, options):
        """Print diff for all conflicts"""
        for conflict in self.branch_builder.conflicts.get():
            if conflict.diff:
                print(conflict)
                print("-" * len(str(conflict)))
                print(conflict.diff)
                print()

    @parse_command
    def do_auto(self, options):
        """Automatically solve conflicts when no feedback from user is necessary."""
        self.branch_builder.manual_resolution = False
        self.branch_builder.resolve_conflicts()

        return True

    @parse_command
    def do_name(self, options):
        """Change the name of the experiment"""
        self.branch_builder.change_experiment_name(options.experiment_name)

        print('TIP: You can use the \'-b\' or \'--branch\' command-line '
              'argument to automate the naming process.')

        self.do_status("")

    @wrap_autocomplete
    def complete_name(self, text, line, begidx, endidx):
        """Auto-completion of experiment name based on names in the EVC three."""
        if len(line.split(" ")) >= 3:
            names = []
        else:
            query = {'refers.root_id': self.branch_builder.experiment_config['refers']['root_id'],
                     'metadata.user': self.branch_builder.experiment_config['metadata']['user']}
            names = [experiment['name'] for experiment in get_storage().fetch_experiments(query)]

        return self._get_completions(names, text)

    def _get_completions(self, names, text, ignore=()):
        return [f + ' ' for f in names if f.startswith(text) and f not in ignore]

    @parse_command
    def do_code(self, options):
        """Set the type of the code change"""
        self.branch_builder.set_code_change_type(options.change_type)

        self.do_status("")

    @wrap_autocomplete
    def complete_code(self, text, line, begidx, endidx):
        """Auto-complete based on possible code change types"""
        names = adapters.CodeChange.types
        if len(line.split(" ")) > 2:
            names = []
        return self._get_completions(names, text)

    @parse_command
    def do_commandline(self, options):
        """Set the type of the commandline change"""
        self.branch_builder.set_cli_change_type(options.change_type)
        self.do_status("")

    @wrap_autocomplete
    def complete_commandline(self, text, line, begidx, endidx):
        """Auto-complete based on possible commandline change types"""
        names = adapters.CommandLineChange.types
        if len(line.split(" ")) > 2:
            names = []
        return self._get_completions(names, text)

    @parse_command
    def do_config(self, options):
        """Set the type of the commandline change"""
        self.branch_builder.set_script_config_change_type(options.change_type)
        self.do_status("")

    @wrap_autocomplete
    def complete_config(self, text, line, begidx, endidx):
        """Auto-complete based on possible script config change types"""
        names = adapters.ScriptConfigChange.types
        if len(line.split(" ")) > 2:
            names = []
        return self._get_completions(names, text)

    @parse_command
    def do_algo(self, options):
        """Resolve the algorithm conflict"""
        self.branch_builder.set_algo()
        self.do_status("")

    @parse_command
    def do_add(self, options):
        """Add the given `new` or `changed` dimension to the configuration"""
        print('TIP: You can use the \'~+\' marker in place of the usual ~ with '
              'the command-line to solve this conflict automatically.'
              '\nEx: -x~+uniform(0,1)')

        self.branch_builder.add_dimension(options.dimension_name, options.default_value)

        self.do_status("")

    @wrap_autocomplete
    def complete_add(self, text, line, begidx, endidx):
        """Auto-complete addition of new or changed dimensions"""
        return self._complete_dim(
            [conflicts.NewDimensionConflict, conflicts.ChangedDimensionConflict],
            text, line, begidx, endidx)

    def _complete_add_new_or_missing(self, conflict, text, line, begidx, endidx):
        """Auto-complete addition of new dimensions or deletion of missing dimensions"""
        dim_is_categorical = conflict.dimension.type == 'categorical'

        if len(line.split(" ")) == 3:
            return ['--default-value ']
        elif len(line.split(" ")) == 4 and dim_is_categorical:
            return self._get_completions([repr(f) for f in conflict.dimension.categories], text)

        return []

    def _complete_add_changed(self, conflict, text, line, begidx, endidx):
        """Auto-complete addition of changed dimensions"""
        return []

    @parse_command
    def do_remove(self, options):
        """Remove the given `missing` dimension from the configuration"""
        print('TIP: You can use the \'~-\' marker in place of the usual ~ with '
              'the command-line to solve this conflict automatically.'
              '\nEx: -x~-')

        self.branch_builder.remove_dimension(options.dimension_name, options.default_value)

        self.do_status("")

    @wrap_autocomplete
    def complete_remove(self, text, line, begidx, endidx):
        """Auto-complete deletion of missing dimensions"""
        return self._complete_dim([conflicts.MissingDimensionConflict], text, line, begidx, endidx)

    def _complete_dim(self, conflict_types, text, line, begidx, endidx):
        """Auto-complete addition or deletion of new, changed or missing dimensions"""
        potential_conflicts = self.branch_builder.conflicts.get_remaining(conflict_types)

        names = [conflict.dimension.name.lstrip("/") for conflict in potential_conflicts]

        if len(line.split(" ")) == 2:
            return self._get_completions(names, text)

        # Dimension name is fully specified in command line
        names = [name for name in names if name == line.split(" ")[1]]
        if not names:
            return []
        name = names[0]

        potential_conflicts = [conflict for conflict in potential_conflicts
                               if conflict.dimension.name.lstrip("/") == name]
        if not potential_conflicts:
            return []
        conflict = potential_conflicts[0]

        if isinstance(conflict,
                      (conflicts.NewDimensionConflict, conflicts.MissingDimensionConflict)):
            return self._complete_add_new_or_missing(conflict, text, line, begidx, endidx)
        else:
            return self._complete_add_changed(conflict, text, line, begidx, endidx)

    @parse_command
    def do_rename(self, options):
        """Rename a dimension"""
        print('TIP: You can use the \'~>\' marker in place of the usual ~ with '
              'the command-line to solve this dimension automatically.'
              '\nEx: -x~>y')
        self.branch_builder.rename_dimension(options.old_name, options.new_name)

        self.do_status("")

    @wrap_autocomplete
    def complete_rename(self, text, line, begidx, endidx):
        """Auto-complete rename based on non-resolved missing dimensions and new dimensions
        conflicts
        """
        if len(line.split(" ")) < 3:
            potential_conflicts = self.branch_builder.conflicts.get_remaining(
                [conflicts.MissingDimensionConflict])
        elif len(line.split(" ")) == 3:
            potential_conflicts = self.branch_builder.conflicts.get_remaining(
                [conflicts.NewDimensionConflict])
        else:
            potential_conflicts = []

        names = [conflict.dimension.name.lstrip('/') for conflict in potential_conflicts]

        return self._get_completions(names, text)

    @parse_command
    def do_reset(self, options):
        """Revert a resolution and mark the conflict as non-resolved"""
        for resolution in options.resolutions:
            self.branch_builder.reset(resolution.strip("' ").replace('"', "'"))

        self.do_status("")

    @wrap_autocomplete
    def complete_reset(self, text, line, begidx, endidx):
        """Auto-complete reset based on current resolutions"""
        names = list(map(str, self.branch_builder.conflicts.get_resolutions()))
        names = [shlex.quote(name) for name in names]
        try:
            ignore = [shlex.quote(t) for t in shlex.split(line)]
        except ValueError:
            ignore = [shlex.quote(t) for t in shlex.split(line + "'")]

        names = [name.replace('\'"\'"\'', '"') for name in names]

        return self._get_completions(names, text, ignore=ignore)

    @parse_command
    def do_abort(self, options):
        """Exit the prompt without saving"""
        print('Closing interactive conflicts solver. Experiment branch won\'t be saved.')

        self.abort = True
        return True

    @parse_command
    def do_quit(self, options):
        """Exit the prompt without saving"""
        return self.do_abort("")

    @parse_command
    def do_q(self, options):
        """Alias for quit"""
        return self.do_quit("")

    @parse_command
    def do_commit(self, options):
        """Exit the prompt and creates the adapters inside the builders"""
        if not self.branch_builder.is_resolved:
            print('Error: There is still remaining issues. Enter \'abort\' or \'quit\' to leave '
                  'without registering the new experiment branch.')
            return False

        new_name = self.branch_builder.conflicting_config['name']
        print('Registering experiment branch \'{0}\'.'.format(new_name))

        return True
